# -*- coding: utf-8 -*-

"""Top-level package for mlbench_core."""

__version__ = "2.4.0-dev4"

from . import api
from . import controlflow
from . import dataset
from . import evaluation
from . import lr_scheduler
from . import models
from . import optim
from . import utils



r"""Learning rate scheduling in tensorflow.

The manual_stepping function is taken from :

https://github.com/tensorflow/models/blob/master/research/object_detection/utils/learning_schedules.py
"""

import tensorflow as tf


def manual_stepping(global_step, boundaries, rates, warmup=False):
    """Manually stepped learning rate schedule.

    This function provides fine grained control over learning rates.  One must
    specify a sequence of learning rates as well as a set of integer steps
    at which the current learning rate must transition to the next.  For example,
    if boundaries = [5, 10] and rates = [.1, .01, .001], then the learning
    rate returned by this function is .1 for global_step=0,...,4, .01 for
    global_step=5...9, and .001 for global_step=10 and onward.

    Args:
        global_step (:obj:`tf.Tensor`): int64 (scalar) tensor representing global step.
        boundaries (list): a list of global steps at which to switch learning
        rates (list): a list of (float) learning rates corresponding to intervals between
            the boundaries.  The length of this list must be exactly len(boundaries) + 1.
        warmup (bool, optional): Defaults to False. Whether to linearly interpolate learning
            rate for steps in [0, boundaries[0]].

    Raises:
        ValueError: boundaries is a strictly increasing list of positive integers
        ValueError: len(rates) == len(boundaries) + 1
        ValueError: boundaries[0] != 0

    Returns:
        :obj:`tf.Tensor`: a (scalar) float tensor representing learning rate
    """

    if any([b < 0 for b in boundaries]) or any(
        [not isinstance(b, int) for b in boundaries]
    ):
        raise ValueError("boundaries must be a list of positive integers")
    if any([bnext <= b for bnext, b in zip(boundaries[1:], boundaries[:-1])]):
        raise ValueError("Entries in boundaries must be strictly increasing.")
    if any([not isinstance(r, float) for r in rates]):
        raise ValueError("Learning rates must be floats")
    if len(rates) != len(boundaries) + 1:
        raise ValueError(
            "Number of provided learning rates must exceed "
            "number of boundary points by exactly 1."
        )

    if boundaries and boundaries[0] == 0:
        raise ValueError("First step cannot be zero.")

    if warmup and boundaries:
        slope = (rates[1] - rates[0]) * 1.0 / boundaries[0]
        warmup_steps = range(boundaries[0])
        warmup_rates = [rates[0] + slope * step for step in warmup_steps]
        boundaries = warmup_steps + boundaries
        rates = warmup_rates + rates[1:]
    else:
        boundaries = [0] + boundaries
    num_boundaries = len(boundaries)
    rate_index = tf.reduce_max(
        tf.where(
            tf.greater_equal(global_step, boundaries),
            list(range(num_boundaries)),
            [0] * num_boundaries,
        )
    )
    return tf.reduce_sum(
        rates * tf.one_hot(rate_index, depth=num_boundaries), name="learning_rate"
    )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class BahdanauAttention(nn.Module):
    """
    Bahdanau Attention (https://arxiv.org/abs/1409.0473)
    Implementation is very similar to tf.contrib.seq2seq.BahdanauAttention
    """

    def __init__(
        self,
        query_size,
        key_size,
        num_units,
        normalize=False,
        batch_first=False,
        init_weight=0.1,
    ):
        """
        Constructor for the BahdanauAttention.

        Args:
            query_size: feature dimension for query
            key_size: feature dimension for keys
            num_units: internal feature dimension
            normalize: whether to normalize energy term
            batch_first: if True batch size is the 1st dimension, if False
                the sequence is first and batch size is second
            init_weight: range for uniform initializer used to initialize
                Linear key and query transform layers and linear_att vector
        """
        super(BahdanauAttention, self).__init__()

        self.normalize = normalize
        self.batch_first = batch_first
        self.num_units = num_units

        self.linear_q = nn.Linear(query_size, num_units, bias=False)
        self.linear_k = nn.Linear(key_size, num_units, bias=False)
        nn.init.uniform_(self.linear_q.weight.data, -init_weight, init_weight)
        nn.init.uniform_(self.linear_k.weight.data, -init_weight, init_weight)

        self.linear_att = Parameter(torch.Tensor(num_units))

        self.mask = None

        if self.normalize:
            self.normalize_scalar = Parameter(torch.Tensor(1))
            self.normalize_bias = Parameter(torch.Tensor(num_units))
        else:
            self.register_parameter("normalize_scalar", None)
            self.register_parameter("normalize_bias", None)

        self.reset_parameters(init_weight)

    def reset_parameters(self, init_weight):
        """
        Sets initial random values for trainable parameters.
        """
        stdv = 1.0 / math.sqrt(self.num_units)
        self.linear_att.data.uniform_(-init_weight, init_weight)

        if self.normalize:
            self.normalize_scalar.data.fill_(stdv)
            self.normalize_bias.data.zero_()

    def set_mask(self, context_len, context):
        """
        Sets self.mask which is applied before softmax
        ones for inactive context fields, zeros for active context fields

        Args:
            context_len:
            context: if batch_first: (b x t_k x n) else: (t_k x b x n)

        Returns:

        """

        if self.batch_first:
            max_len = context.size(1)
        else:
            max_len = context.size(0)

        indices = torch.arange(0, max_len, dtype=torch.int64, device=context.device)
        self.mask = indices >= (context_len.unsqueeze(1))

    def calc_score(self, att_query, att_keys):
        """
        Calculate Bahdanau score

        Args:
            att_query:
            att_keys:

        Returns:

        """

        b, t_k, n = att_keys.size()
        t_q = att_query.size(1)

        att_query = att_query.unsqueeze(2).expand(b, t_q, t_k, n)
        att_keys = att_keys.unsqueeze(1).expand(b, t_q, t_k, n)
        sum_qk = att_query + att_keys

        if self.normalize:
            sum_qk = sum_qk + self.normalize_bias
            linear_att = self.linear_att / self.linear_att.norm()
            linear_att = linear_att * self.normalize_scalar
        else:
            linear_att = self.linear_att

        out = torch.tanh(sum_qk).matmul(linear_att)
        return out

    def forward(self, query, keys):
        """

        Args:
            query: if batch_first: (b x t_q x n) else: (t_q x b x n)
            keys: if batch_first: (b x t_k x n) else (t_k x b x n)

        Returns:
            (context, scores_normalized)
        context: if batch_first: (b x t_q x n) else (t_q x b x n)
        scores_normalized: if batch_first (b x t_q x t_k) else (t_q x b x t_k)

        """

        # first dim of keys and query has to be 'batch', it's needed for bmm
        if not self.batch_first:
            keys = keys.transpose(0, 1)
            if query.dim() == 3:
                query = query.transpose(0, 1)

        if query.dim() == 2:
            single_query = True
            query = query.unsqueeze(1)
        else:
            single_query = False

        b = query.size(0)
        t_k = keys.size(1)
        t_q = query.size(1)

        # FC layers to transform query and key
        processed_query = self.linear_q(query)
        processed_key = self.linear_k(keys)

        # scores: (b x t_q x t_k)
        scores = self.calc_score(processed_query, processed_key)

        if self.mask is not None:
            mask = self.mask.unsqueeze(1).expand(b, t_q, t_k)
            # I can't use -INF because of overflow check in pytorch
            scores.data.masked_fill_(mask, -65504.0)

        # Normalize the scores, softmax over t_k
        scores_normalized = F.softmax(scores, dim=-1)

        # Calculate the weighted average of the attention inputs according to
        # the scores
        # context: (b x t_q x n)
        context = torch.bmm(scores_normalized, keys)

        if single_query:
            context = context.squeeze(1)
            scores_normalized = scores_normalized.squeeze(1)
        elif not self.batch_first:
            context = context.transpose(0, 1)
            scores_normalized = scores_normalized.transpose(0, 1)

        return context, scores_normalized



import mlbench_core.dataset.translation.pytorch.config as config
import torch.nn as nn
from mlbench_core.models.pytorch.gnmt.utils import init_lstm_
from torch.nn.utils.rnn import pack_padded_sequence
from torch.nn.utils.rnn import pad_packed_sequence


class ResidualRecurrentEncoder(nn.Module):
    """
    Encoder with Embedding, LSTM layers, residual connections and optional
    dropout.

    The first LSTM layer is bidirectional and uses variable sequence length
    API, the remaining (num_layers-1) layers are unidirectional. Residual
    connections are enabled after third LSTM layer, dropout is applied on
    inputs to LSTM layers.
    """

    def __init__(
        self,
        vocab_size,
        hidden_size=1024,
        num_layers=4,
        dropout=0.2,
        batch_first=False,
        embedder=None,
        init_weight=0.1,
    ):
        """
        Constructor for the ResidualRecurrentEncoder.

        Args:
            vocab_size: size of vocabulary
            hidden_size: hidden size for LSTM layers
            num_layers: number of LSTM layers, 1st layer is bidirectional
            dropout: probability of dropout (on input to LSTM layers)
            batch_first: if True the model uses (batch,seq,feature) tensors,
                if false the model uses (seq, batch, feature)
            embedder: instance of nn.Embedding, if None constructor will
                create new embedding layer
            init_weight: range for the uniform initializer
        """
        super(ResidualRecurrentEncoder, self).__init__()
        self.batch_first = batch_first
        self.rnn_layers = nn.ModuleList()
        # 1st LSTM layer, bidirectional
        self.rnn_layers.append(
            nn.LSTM(
                hidden_size,
                hidden_size,
                num_layers=1,
                bias=True,
                batch_first=batch_first,
                bidirectional=True,
            )
        )

        # 2nd LSTM layer, with 2x larger input_size
        self.rnn_layers.append(
            nn.LSTM(
                (2 * hidden_size),
                hidden_size,
                num_layers=1,
                bias=True,
                batch_first=batch_first,
            )
        )

        # Remaining LSTM layers
        for _ in range(num_layers - 2):
            self.rnn_layers.append(
                nn.LSTM(
                    hidden_size,
                    hidden_size,
                    num_layers=1,
                    bias=True,
                    batch_first=batch_first,
                )
            )

        for lstm in self.rnn_layers:
            init_lstm_(lstm, init_weight)

        self.dropout = nn.Dropout(p=dropout)

        if embedder is not None:
            self.embedder = embedder
        else:
            self.embedder = nn.Embedding(
                vocab_size, hidden_size, padding_idx=config.PAD
            )
            nn.init.uniform_(self.embedder.weight.data, -init_weight, init_weight)

    def forward(self, inputs, lengths):
        """
        Execute the encoder.

        Args:
            inputs: tensor with indices from the vocabulary
            lengths: vector with sequence lengths (excluding padding)

        Returns:
            tensor with encoded sequences

        """
        x = self.embedder(inputs)

        # bidirectional layer
        x = self.dropout(x)
        x = pack_padded_sequence(x, lengths, batch_first=self.batch_first)
        x, _ = self.rnn_layers[0](x)
        x, _ = pad_packed_sequence(x, batch_first=self.batch_first)

        # 1st unidirectional layer
        x = self.dropout(x)
        x, _ = self.rnn_layers[1](x)

        # the rest of unidirectional layers,
        # with residual connections starting from 3rd layer
        for i in range(2, len(self.rnn_layers)):
            residual = x
            x = self.dropout(x)
            x, _ = self.rnn_layers[i](x)
            x = x + residual

        return x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Imports
#

import sys
import os

from os.path import abspath, join, dirname

sys.path.insert(0, abspath(join(dirname(__file__))))
sys.path.insert(0, abspath(join(dirname(__file__), "..")))

import sys
from unittest.mock import MagicMock


class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


MOCK_MODULES = [
    "torch",
    "torch.distributed",
    "torch.nn",
    "torch.nn.functional",
    "torch.nn.modules",
    "torch.nn.modules.loss",
    "torch.utils",
    "torch.utils.model_zoo",
    "torch.nn.init",
    "torch.utils.data",
    "torch.optim",
    "torch.optim.lr_scheduler",
    "torchvision",
    "torchvision.datasets",
    "torchvision.transforms",
]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

import mlbench_core


# -- RTD configuration ------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# This is used for linking and such so we link to the thing we're building
rtd_version = os.environ.get("READTHEDOCS_VERSION", "latest")
if rtd_version not in ["stable", "latest"]:
    rtd_version = "stable"

# -- General configuration ------------------------------------------------


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinxcontrib.napoleon",
    "sphinxcontrib.bibtex",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
html_static_path = ["_static"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

autodoc_mock_imports = ["torch"]

# General information about the project.
project = "MLBench Core"
copyright = "2018 MLBench development team"

# napoleon
napoleon_google_docstring = True
napoleon_numpy_docstring = False


autoclass_content = "both"

intersphinx_mapping = {}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
    print(html_theme_path)
else:
    html_theme = "default"

html_theme_options = {"navigation_depth": 5}

# Output file base name for HTML help builder.
htmlhelp_basename = "MLBench_Core"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        "index",
        "MLBench_DaMLBench_Coreshboard.tex",
        "MLBench Core Documentation",
        "MLBench development team",
        "manual",
    ),
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "MLBench_Core",
        "MLBench Core Documentation",
        "MLBench Core development team",
        "MLBench_Core",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "MLBench Core"
epub_author = "MLBench development team"
epub_publisher = "MLBench development team"
epub_copyright = "2018, MLBench development team"

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Custom Document processing ----------------------------------------------

import gensidebar

gensidebar.generate_sidebar(globals(), "mlbench_core")

import sphinx.addnodes
import docutils.nodes


def process_child(node):
    """This function changes class references to not have the
       intermediate module name by hacking at the doctree"""

    # Edit descriptions to be nicer
    if isinstance(node, sphinx.addnodes.desc_addname):
        if len(node.children) == 1:
            child = node.children[0]
            text = child.astext()

    # Edit literals to be nicer
    elif isinstance(node, docutils.nodes.literal):
        child = node.children[0]
        text = child.astext()

    for child in node.children:
        process_child(child)


def doctree_read(app, doctree):
    for child in doctree.children:
        process_child(child)


def setup(app):
    print("Yay!")
    app.add_css_file("css/custom.css")
    app.connect("doctree-read", doctree_read)



r"""Partition PyTorch datasets."""
# -*- coding: utf-8 -*-
import random
import logging

import numpy as np
import torch
import torch.distributed as dist

from mlbench_core.utils.pytorch.distributed import get_backend_tensor

_logger = logging.getLogger("mlbench")


class Partition(object):
    """Dataset-like object, but only access a subset of it.

    Wraps a dataset, only exposing the entries selected by the `indices`
    parameter.

    Args:
        data (:obj:`list` of data entries): The data to partition over
        indices (:obj:`list` of :obj:`int`): indices of entries to use
    """

    def __init__(self, data, indices):
        self.data = data
        self.indices = indices

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index):
        data_idx = self.indices[index]
        return self.data[data_idx]

    def __getattr__(self, item):
        return self.data.__dict__[item]


class Partitioner(object):
    """Use a partition of dataset."""

    def consistent_indices(self, rank, indices, shuffle):
        r""" synchronize indices among workers. """
        if rank == 0 and shuffle:
            random.shuffle(indices)

        # broadcast.
        indices = get_backend_tensor(torch.IntTensor(indices))

        dist.broadcast(indices, src=0)

        return list(indices)


class DataPartitioner(Partitioner):
    """ Partitions a dataset into different sized chunks.

    Used for train:test:validation split.

    Args:
        data (:obj:`list` of data entries): The data to partition over
        rank (int): The rank of the current node
        shuffle (bool): Whether to shuffle entries or not
        sizes (:obj:`list` of :obj:`float`): The relative sizes of the
            splits. Should sum up to 1.0. (Default = [0.7, 0.2, 0.1])
    """

    def __init__(self, data, rank, shuffle, sizes=[0.7, 0.2, 0.1]):
        # prepare info.
        self.data = data
        self.data_size = len(self.data)
        self.partitions = []

        # get shuffled/unshuffled data.
        indices = [x for x in range(0, self.data_size)]
        indices = self.consistent_indices(rank, indices, shuffle)

        # partition indices.
        sizes = np.cumsum(sizes)
        from_index = 0
        for ind, _ in enumerate(sizes):
            to_index = int(sizes[ind] * self.data_size)
            self.partitions.append(indices[from_index:to_index])
            from_index = to_index

    def use(self, partition_ind):
        """Return a partition of data.

        Args:
            partition_ind (int): The index of the partition to get
        """
        return Partition(self.data, self.partitions[partition_ind])


def partition_dataset_by_rank(
    dataset, rank, world_size, distribution="uniform", shuffle=True
):
    r"""Given a dataset, partition it by a distribution and each rank takes part of data.

    Args:
        dataset (:obj:`torch.utils.data.Dataset`): The dataset
        rank (int): The rank of the current worker
        world_size (int): The total number of workers
        distribution (str): The sampling distribution to use. Default: `uniform`
        shuffle (bool): Whether to shuffle the dataset before partitioning. Default: `True`
    """
    if distribution != "uniform":
        raise NotImplementedError(
            "Distribution {} not implemented.".format(distribution)
        )

    partition_sizes = [1.0 / world_size for _ in range(world_size)]
    partition = DataPartitioner(dataset, rank, shuffle, partition_sizes)
    partitioned_data = partition.use(rank)
    _logger.debug("Partition dataset use {}-th.".format(rank))
    return partitioned_data



"""
Utility functions that allow the download and transformation of LIBSVM datasets
into LMDB format.

https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary.html

"""
import os

import click
from mlbench_core.dataset.util.tools import maybe_download_and_extract_bz2
from sklearn.datasets import load_svmlight_file
from sklearn.datasets import make_classification
from tensorpack.dataflow import PrefetchDataZMQ, LMDBSerializer

_DATASET_MAP = {
    "australian_train": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools"
        "/datasets/binary/australian",
        "sparse": False,
    },
    "duke_train": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets"
        "/binary/duke.tr.bz2",
        "sparse": True,
    },
    "duke_test": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets"
        "/binary/duke.val.bz2",
        "sparse": True,
    },
    "epsilon_train": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools"
        "/datasets/binary/epsilon_normalized.bz2",
        "sparse": False,
    },
    "epsilon_test": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets"
        "/binary/epsilon_normalized.t.bz2",
        "sparse": False,
    },
    "rcv1_train": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets"
        "/binary/rcv1_train.binary.bz2",
        "sparse": True,
    },
    "rcv1_test": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets"
        "/binary/rcv1_test.binary.bz2",
        "sparse": True,
    },
    "webspam_train": {
        "url": "https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary"
        "/webspam_wc_normalized_trigram.svm.bz2",
        "sparse": True,
    },
}


def _get_dense_tensor(tensor):
    if "sparse" in str(type(tensor)):
        return tensor.toarray()
    elif "numpy" in str(type(tensor)):
        return tensor


def _correct_binary_labels(labels, is_01_classes=True):
    classes = set(labels)

    if -1 in classes and is_01_classes:
        labels[labels == -1] = 0
    return labels


def _get_features_and_labels(data, is_sparse):
    features, labels = data

    features = _get_dense_tensor(features) if not is_sparse else features
    labels = _get_dense_tensor(labels)
    labels = _correct_binary_labels(labels)
    return features, labels


def _load_libsvm_data(root_path, name, data_type):
    data = _DATASET_MAP["{}_{}".format(name, data_type)]
    data_url = data["url"]
    is_sparse = data["sparse"]

    file_name = data_url.split("/")[-1]
    # Downloads and extracts the data (deletes downloaded file after)
    file_path = maybe_download_and_extract_bz2(root_path, file_name, data_url)

    print("Loading SVM file {}".format(file_path))
    data = load_svmlight_file(file_path)
    features, labels = _get_features_and_labels(data, is_sparse)
    return features, labels, is_sparse


class LIBSVMDataset(object):
    def __init__(self, features, labels, is_sparse):

        self.is_sparse = is_sparse

        self.features, self.labels = features, labels

    def __len__(self):
        return self.features.shape[0]

    def __iter__(self):
        idxs = list(range(self.__len__()))
        for k in idxs:
            features = self.features[k]
            label = self.labels[k]
            if self.is_sparse:
                features = features.tocoo()
                yield [(features.data, features.row, features.col), label]
            else:
                yield [features, label]

    def get_data(self):
        return self.__iter__()

    def size(self):
        return self.__len__()

    def reset_state(self):
        """
        Reset state of the dataflow.
        It **has to** be called once and only once before producing datapoints.
        Note:
            1. If the dataflow is forked, each process will call this method
               before producing datapoints.
            2. The caller thread of this method must remain alive to keep
            this dataflow alive.
        For example, RNG **has to** be reset if used in the DataFlow,
        otherwise it won't work well with prefetching, because different
        processes will have the same RNG state.
        """
        pass


def sequential_epsilon_or_rcv1(root_path, name, data_type):
    features, labels, is_sparse = _load_libsvm_data(root_path, name, data_type)
    data = LIBSVMDataset(features, labels, is_sparse)
    lmdb_file_path = os.path.join(root_path, "{}_{}.lmdb".format(name, data_type))

    ds1 = PrefetchDataZMQ(data, nr_proc=1)
    LMDBSerializer.save(ds1, lmdb_file_path)

    print("Dumped dataflow to {} for {}".format(lmdb_file_path, name))


def sequential_synthetic_dataset(root_path, dataset_name, data_type):
    """Generate a synthetic dataset for regression."""
    if data_type == "dense":
        X, y = make_classification(
            n_samples=10000,
            n_features=100,
            n_informative=90,
            n_classes=2,
            random_state=42,
        )
    else:
        raise NotImplementedError(
            "{} synthetic dataset is " "not supported.".format(data_type)
        )

    data = LIBSVMDataset(X, y, False)
    lmdb_file_path = os.path.join(
        root_path, "{}_{}.lmdb".format(dataset_name, data_type)
    )

    ds1 = PrefetchDataZMQ(data, nr_proc=1)
    LMDBSerializer.save(ds1, lmdb_file_path)

    print("Dumped dataflow to {} for {}".format(lmdb_file_path, dataset_name))


@click.command()
@click.argument("data")
@click.argument("data_type")
@click.argument("data_dir")
def generate_lmdb_from_libsvm(data, data_type, data_dir):
    """ Utility script that downloads data from LIBSVM and transforms them
    into `.lmdb` files.

    Args:
        data (str): Name of dataset
        data_type (str): One of `train` `test`
        data_dir (str): Directory where to download, extract and transform
    """
    if "epsilon" in data or "rcv1" in data or "webspam" in data:
        sequential_epsilon_or_rcv1(data_dir, data, data_type)
    elif "australian" in data or "duke" in data:
        # These two are small datasets for testing purpose.
        sequential_epsilon_or_rcv1(data_dir, data, data_type)
    elif "synthetic" in data:
        sequential_synthetic_dataset(data_dir, data, data_type)
    else:
        raise NotImplementedError("Dataset {} not supported.".format(data))


if __name__ == "__main__":
    generate_lmdb_from_libsvm()



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains definitions for Residual Networks.
Residual networks ('v1' ResNets) were originally proposed in :cite:`he2016deep`.
The full preactivation 'v2' ResNet variant was introduced by :cite:`he2016identity`.
The key difference of the full preactivation 'v2' variant compared to the
'v1' variant in [1] is the use of batch normalization before every weight layer
rather than after.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

# _BATCH_NORM_DECAY = 1.0
_BATCH_NORM_DECAY = 0.9
_BATCH_NORM_EPSILON = 1e-5
DEFAULT_VERSION = 2
DEFAULT_DTYPE = tf.float32
CASTABLE_TYPES = (tf.float16,)
ALLOWED_TYPES = (DEFAULT_DTYPE,) + CASTABLE_TYPES

################################################################################
# Convenience functions for building the ResNet model.
################################################################################


def batch_norm(inputs, training, data_format):
    """Performs a batch normalization using a standard set of parameters."""
    # We set fused=True for a significant performance boost. See
    # https://www.tensorflow.org/performance/performance_guide#common_fused_ops
    return tf.layers.batch_normalization(
        inputs=inputs,
        axis=1 if data_format == "channels_first" else 3,
        momentum=_BATCH_NORM_DECAY,
        epsilon=_BATCH_NORM_EPSILON,
        center=True,
        scale=True,
        trainable=True,
        training=training,
        fused=True,
    )


def fixed_padding(inputs, kernel_size, data_format):
    """Pads the input along the spatial dimensions independently of input size.

    Args:
      inputs (:obj:`tf.Tensor`): A tensor of size [batch, channels, height_in, width_in] or
        [batch, height_in, width_in, channels] depending on data_format.
      kernel_size (int): The kernel to be used in the conv2d or max_pool2d operation.
                   Should be a positive integer.
      data_format (str): The input format ('channels_last' or 'channels_first').

    Returns:
      A tensor with the same format as the input with the data either intact
      (if kernel_size == 1) or padded (if kernel_size > 1).
    """
    pad_total = kernel_size - 1
    pad_beg = pad_total // 2
    pad_end = pad_total - pad_beg

    if data_format == "channels_first":
        padded_inputs = tf.pad(
            inputs, [[0, 0], [0, 0], [pad_beg, pad_end], [pad_beg, pad_end]]
        )
    else:
        padded_inputs = tf.pad(
            inputs, [[0, 0], [pad_beg, pad_end], [pad_beg, pad_end], [0, 0]]
        )
    return padded_inputs


def conv2d_fixed_padding(inputs, filters, kernel_size, strides, data_format):
    """Strided 2-D convolution with explicit padding."""
    # The padding is consistent and is based only on `kernel_size`, not on the
    # dimensions of `inputs` (as opposed to using `tf.layers.conv2d` alone).
    with tf.name_scope("conv2d_fixed_padding"):
        tf.identity(inputs, name="inputs_pre")
        if strides > 1:
            inputs = fixed_padding(inputs, kernel_size, data_format)
        tf.identity(inputs, name="inputs")

        layer = tf.layers.conv2d(
            inputs=inputs,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=("SAME" if strides == 1 else "VALID"),
            use_bias=False,
            kernel_initializer=tf.variance_scaling_initializer(),
            data_format=data_format,  # name="conv2d"
        )

        return layer


################################################################################
# ResNet block definitions.
################################################################################
def _building_block_v1(
    inputs, filters, training, projection_shortcut, strides, data_format
):
    """A single block for ResNet v1, without a bottleneck.

    Convolution then batch normalization then ReLU as described by:
      Deep Residual Learning for Image Recognition
      https://arxiv.org/pdf/1512.03385.pdf
      by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Dec 2015.

    Args:
      inputs (:obj:`tf.Tensor`): A tensor of size [batch, channels, height_in, width_in] or
        [batch, height_in, width_in, channels] depending on data_format.
      filters (int): The number of filters for the convolutions.
      training (bool): A Boolean for whether the model is in training or inference
        mode. Needed for batch normalization.
      projection_shortcut (callable): The function to use for projection shortcuts
        (typically a 1x1 convolution when downsampling the input).
      strides (int): The block's stride. If greater than 1, this block will ultimately
        downsample the input.
      data_format (str): The input format ('channels_last' or 'channels_first').

    Returns:
      The output tensor of the block; shape should match inputs.
    """
    shortcut = inputs

    if projection_shortcut is not None:
        shortcut = projection_shortcut(inputs)
        shortcut = batch_norm(
            inputs=shortcut, training=training, data_format=data_format
        )

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=3,
        strides=strides,
        data_format=data_format,
    )
    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=3,
        strides=1,
        data_format=data_format,
    )
    inputs = batch_norm(inputs, training, data_format)
    inputs += shortcut
    inputs = tf.nn.relu(inputs)

    return inputs


def _building_block_v2(
    inputs, filters, training, projection_shortcut, strides, data_format
):
    """A single block for ResNet v2, without a bottleneck.

    Batch normalization then ReLu then convolution as described by:
      Identity Mappings in Deep Residual Networks
      https://arxiv.org/pdf/1603.05027.pdf
      by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Jul 2016.

    Args:
      inputs (:obj:`tf.Tensor`): A tensor of size [batch, channels, height_in, width_in] or
        [batch, height_in, width_in, channels] depending on data_format.
      filters (int): The number of filters for the convolutions.
      training (bool): A Boolean for whether the model is in training or inference
        mode. Needed for batch normalization.
      projection_shortcut (callable): The function to use for projection shortcuts
        (typically a 1x1 convolution when downsampling the input).
      strides (int): The block's stride. If greater than 1, this block will ultimately
        downsample the input.
      data_format (str): The input format ('channels_last' or 'channels_first').

    Returns:
      The output tensor of the block; shape should match inputs.
    """
    shortcut = inputs
    tf.identity(shortcut, name="inputs")

    with tf.name_scope("test_batch_norm"):
        inputs = batch_norm(inputs, training, data_format)
    tf.identity(inputs, name="inputs_batch_normed")

    inputs = tf.nn.relu(inputs)
    tf.identity(inputs, name="inputs_batch_normed_relu")

    # The projection shortcut should come after the first batch norm and ReLU
    # since it performs a 1x1 convolution.
    if projection_shortcut is not None:
        shortcut = projection_shortcut(inputs)
        tf.identity(shortcut, name="projected_shortcut")

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=3,
        strides=strides,
        data_format=data_format,
    )

    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)
    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=3,
        strides=1,
        data_format=data_format,
    )

    return inputs + shortcut


def _bottleneck_block_v1(
    inputs, filters, training, projection_shortcut, strides, data_format
):
    """A single block for ResNet v1, with a bottleneck.

    Similar to _building_block_v1(), except using the "bottleneck" blocks
    described in:
      Convolution then batch normalization then ReLU as described by:
        Deep Residual Learning for Image Recognition
        https://arxiv.org/pdf/1512.03385.pdf
        by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Dec 2015.

    Args:
      inputs (:obj:`tf.Tensor`): A tensor of size [batch, channels, height_in, width_in] or
        [batch, height_in, width_in, channels] depending on data_format.
      filters (int): The number of filters for the convolutions.
      training (bool): A Boolean for whether the model is in training or inference
        mode. Needed for batch normalization.
      projection_shortcut (callable): The function to use for projection shortcuts
        (typically a 1x1 convolution when downsampling the input).
      strides (int): The block's stride. If greater than 1, this block will ultimately
        downsample the input.
      data_format (str): The input format ('channels_last' or 'channels_first').

    Returns:
      The output tensor of the block; shape should match inputs.
    """
    shortcut = inputs

    if projection_shortcut is not None:
        shortcut = projection_shortcut(inputs)
        shortcut = batch_norm(
            inputs=shortcut, training=training, data_format=data_format
        )

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=1,
        strides=1,
        data_format=data_format,
    )
    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=3,
        strides=strides,
        data_format=data_format,
    )
    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=4 * filters,
        kernel_size=1,
        strides=1,
        data_format=data_format,
    )
    inputs = batch_norm(inputs, training, data_format)
    inputs += shortcut
    inputs = tf.nn.relu(inputs)

    return inputs


def _bottleneck_block_v2(
    inputs, filters, training, projection_shortcut, strides, data_format
):
    """A single block for ResNet v2, with a bottleneck.

    Similar to _building_block_v2(), except using the "bottleneck" blocks
    described in:
      Convolution then batch normalization then ReLU as described by:
        Deep Residual Learning for Image Recognition
        https://arxiv.org/pdf/1512.03385.pdf
        by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Dec 2015.
    Adapted to the ordering conventions of:
      Batch normalization then ReLu then convolution as described by:
        Identity Mappings in Deep Residual Networks
        https://arxiv.org/pdf/1603.05027.pdf
        by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Jul 2016.

    Args:
      inputs (:obj:`tf.Tensor`): A tensor of size [batch, channels, height_in, width_in] or
        [batch, height_in, width_in, channels] depending on data_format.
      filters (int): The number of filters for the convolutions.
      training (bool): A Boolean for whether the model is in training or inference
        mode. Needed for batch normalization.
      projection_shortcut (callable): The function to use for projection shortcuts
        (typically a 1x1 convolution when downsampling the input).
      strides (int): The block's stride. If greater than 1, this block will ultimately
        downsample the input.
      data_format (str): The input format ('channels_last' or 'channels_first').

    Returns:
      The output tensor of the block; shape should match inputs.
    """
    shortcut = inputs
    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)

    # The projection shortcut should come after the first batch norm and ReLU
    # since it performs a 1x1 convolution.
    if projection_shortcut is not None:
        shortcut = projection_shortcut(inputs)

    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=1,
        strides=1,
        data_format=data_format,
    )

    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)
    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=filters,
        kernel_size=3,
        strides=strides,
        data_format=data_format,
    )

    inputs = batch_norm(inputs, training, data_format)
    inputs = tf.nn.relu(inputs)
    inputs = conv2d_fixed_padding(
        inputs=inputs,
        filters=4 * filters,
        kernel_size=1,
        strides=1,
        data_format=data_format,
    )

    return inputs + shortcut


def block_layer(
    inputs, filters, bottleneck, block_fn, blocks, strides, training, name, data_format
):
    """Creates one layer of blocks for the ResNet model.

    Args:
      inputs (:obj:`tf.Tensor`): A tensor of size [batch, channels, height_in, width_in] or
        [batch, height_in, width_in, channels] depending on data_format.
      filters (int): The number of filters for the first convolution of the layer.
      bottleneck (bool): Is the block created a bottleneck block.
      block_fn (callable): The block to use within the model, either `building_block` or
        `bottleneck_block`.
      blocks (int): The number of blocks contained in the layer.
      strides (int): The stride to use for the first convolution of the layer. If
        greater than 1, this layer will ultimately downsample the input.
      training (bool): Either True or False, whether we are currently training the
        model. Needed for batch norm.
      name (str): A string name for the tensor output of the block layer.
      data_format (str): The input format ('channels_last' or 'channels_first').

    Returns:
      The output tensor of the block layer.
    """

    # Bottleneck blocks end with 4x the number of filters as they start with
    filters_out = filters * 4 if bottleneck else filters

    def projection_shortcut(inputs):
        with tf.name_scope("projection_shortcut"):
            layer = conv2d_fixed_padding(
                inputs=inputs,
                filters=filters_out,
                kernel_size=1,
                strides=strides,
                data_format=data_format,
            )
        return layer

    # Only the first block per block_layer uses projection_shortcut and strides
    with tf.name_scope("block0"):
        inputs = block_fn(
            inputs, filters, training, projection_shortcut, strides, data_format
        )

    for i in range(1, blocks):
        with tf.name_scope("block{}".format(i)):
            inputs = block_fn(inputs, filters, training, None, 1, data_format)

    return tf.identity(inputs, name)


class Model(object):
    """Base class for building the Resnet Model."""

    def __init__(
        self,
        resnet_size,
        bottleneck,
        num_classes,
        num_filters,
        kernel_size,
        conv_stride,
        first_pool_size,
        first_pool_stride,
        block_sizes,
        block_strides,
        resnet_version=DEFAULT_VERSION,
        data_format=None,
        dtype=DEFAULT_DTYPE,
    ):
        """Creates a model for classifying an image.

        Args:
          resnet_size (int): A single integer for the size of the ResNet model.
          bottleneck (bool): Use regular blocks or bottleneck blocks.
          num_classes (int): The number of classes used as labels.
          num_filters (int): The number of filters to use for the first block layer
            of the model. This number is then doubled for each subsequent block
            layer.
          kernel_size (int): The kernel size to use for convolution.
          conv_stride (int): stride size for the initial convolutional layer
          first_pool_size (int or None): Pool size to be used for the first pooling layer.
            If none, the first pooling layer is skipped.
          first_pool_stride (int or None): stride size for the first pooling layer. Not used
            if first_pool_size is None.
          block_sizes (list): A list containing n values, where n is the number of sets of
            block layers desired. Each value should be the number of blocks in the
            i-th set.
          block_strides (list): List of integers representing the desired stride size for
            each of the sets of block layers. Should be same length as block_sizes.
          resnet_version (int): Integer representing which version of the ResNet network
            to use. See README for details. Valid values: [1, 2]
          data_format (str): Input format ('channels_last', 'channels_first', or None).
            If set to None, the format is dependent on whether a GPU is available.
          dtype (type): The TensorFlow dtype to use for calculations. If not specified
            tf.float32 is used.

        Raises:
          ValueError: if invalid version is selected.
        """
        self.resnet_size = resnet_size

        if not data_format:
            data_format = (
                "channels_first" if tf.test.is_built_with_cuda() else "channels_last"
            )

        self.resnet_version = resnet_version
        if resnet_version not in (1, 2):
            raise ValueError(
                "Resnet version should be 1 or 2. See README for citations."
            )

        self.bottleneck = bottleneck
        if bottleneck:
            if resnet_version == 1:
                self.block_fn = _bottleneck_block_v1
            else:
                self.block_fn = _bottleneck_block_v2
        else:
            if resnet_version == 1:
                self.block_fn = _building_block_v1
            else:
                self.block_fn = _building_block_v2

        if dtype not in ALLOWED_TYPES:
            raise ValueError("dtype must be one of: {}".format(ALLOWED_TYPES))

        self.data_format = data_format
        self.num_classes = num_classes
        self.num_filters = num_filters
        self.kernel_size = kernel_size
        self.conv_stride = conv_stride
        self.first_pool_size = first_pool_size
        self.first_pool_stride = first_pool_stride
        self.block_sizes = block_sizes
        self.block_strides = block_strides
        self.dtype = dtype
        self.pre_activation = resnet_version == 2

    def _custom_dtype_getter(
        self, getter, name, shape=None, dtype=DEFAULT_DTYPE, *args, **kwargs
    ):
        """Creates variables in fp32, then casts to fp16 if necessary.

        This function is a custom getter. A custom getter is a function with the
        same signature as tf.get_variable, except it has an additional getter
        parameter. Custom getters can be passed as the `custom_getter` parameter of
        tf.variable_scope. Then, tf.get_variable will call the custom getter,
        instead of directly getting a variable itself. This can be used to change
        the types of variables that are retrieved with tf.get_variable.
        The `getter` parameter is the underlying variable getter, that would have
        been called if no custom getter was used. Custom getters typically get a
        variable with `getter`, then modify it in some way.
        This custom getter will create an fp32 variable. If a low precision
        (e.g. float16) variable was requested it will then cast the variable to the
        requested dtype. The reason we do not directly create variables in low
        precision dtypes is that applying small gradients to such variables may
        cause the variable not to change.

        Args:
          getter: The underlying variable getter, that has the same signature as
            tf.get_variable and returns a variable.
          name: The name of the variable to get.
          shape: The shape of the variable to get.
          dtype (type): The dtype of the variable to get. Note that if this is a low
            precision dtype, the variable will be created as a tf.float32 variable,
            then cast to the appropriate dtype
          *args: Additional arguments to pass unmodified to getter.
          **kwargs: Additional keyword arguments to pass unmodified to getter.

        Returns:
          A variable which is cast to fp16 if necessary.
        """

        if dtype in CASTABLE_TYPES:
            var = getter(name, shape, tf.float32, *args, **kwargs)
            return tf.cast(var, dtype=dtype, name=name + "_cast")
        else:
            return getter(name, shape, dtype, *args, **kwargs)

    def _model_variable_scope(self):
        """Returns a variable scope that the model should be created under.

        If self.dtype is a castable type, model variable will be created in fp32
        then cast to self.dtype before being used.

        Returns:
          A variable scope for the model.
        """
        return tf.variable_scope(
            "resnet_model", custom_getter=self._custom_dtype_getter
        )

    def __call__(self, inputs, training):
        """Add operations to classify a batch of input images.

        Args:
          inputs (:obj:`tf.Tensor`): A Tensor representing a batch of input images.
          training (:obj:`tf.Tensor`): A boolean. Set to True to add operations required only when
            training the classifier.

        Returns:
          A logits Tensor with shape [<batch_size>, self.num_classes].
        """

        with self._model_variable_scope():
            if self.data_format == "channels_first":
                # Convert the inputs from channels_last (NHWC) to channels_first (NCHW).
                # This provides a large performance boost on GPU. See
                # https://www.tensorflow.org/performance/performance_guide#data_formats
                inputs = tf.transpose(inputs, [0, 3, 1, 2])

            inputs = conv2d_fixed_padding(
                inputs=inputs,
                filters=self.num_filters,
                kernel_size=self.kernel_size,
                strides=self.conv_stride,
                data_format=self.data_format,
            )
            inputs = tf.identity(inputs, "initial_conv")

            # We do not include batch normalization or activation functions in V2
            # for the initial conv1 because the first ResNet unit will perform these
            # for both the shortcut and non-shortcut paths as part of the first
            # block's projection. Cf. Appendix of [2].
            if self.resnet_version == 1:
                inputs = batch_norm(inputs, training, self.data_format)
                inputs = tf.nn.relu(inputs)

            if self.first_pool_size:
                inputs = tf.layers.max_pooling2d(
                    inputs=inputs,
                    pool_size=self.first_pool_size,
                    strides=self.first_pool_stride,
                    padding="SAME",
                    data_format=self.data_format,
                )
                inputs = tf.identity(inputs, "initial_max_pool")

            for i, num_blocks in enumerate(self.block_sizes):
                num_filters = self.num_filters * (2 ** i)
                with tf.name_scope("block_layer_{}".format(i + 1)):
                    inputs = block_layer(
                        inputs=inputs,
                        filters=num_filters,
                        bottleneck=self.bottleneck,
                        block_fn=self.block_fn,
                        blocks=num_blocks,
                        strides=self.block_strides[i],
                        training=training,
                        name="block_layer{}".format(i + 1),
                        data_format=self.data_format,
                    )

            # Only apply the BN and ReLU for model that does pre_activation in each
            # building/bottleneck block, eg resnet V2.
            if self.pre_activation:
                inputs = batch_norm(inputs, training, self.data_format)
                inputs = tf.nn.relu(inputs)

            # The current top layer has shape
            # `batch_size x pool_size x pool_size x final_size`.
            # ResNet does an Average Pooling layer over pool_size,
            # but that is the same as doing a reduce_mean. We do a reduce_mean
            # here because it performs better than AveragePooling2D.
            axes = [2, 3] if self.data_format == "channels_first" else [1, 2]
            inputs = tf.reduce_mean(inputs, axes, keepdims=True)
            inputs = tf.identity(inputs, "final_reduce_mean")

            inputs = tf.squeeze(inputs, axes)
            inputs = tf.layers.dense(inputs=inputs, units=self.num_classes)
            inputs = tf.identity(inputs, "final_dense")
            return inputs


class Cifar10Model(Model):
    """Model class with appropriate defaults for CIFAR-10 data."""

    def __init__(
        self,
        resnet_size,
        data_format=None,
        num_classes=10,
        resnet_version=2,
        dtype=tf.float32,
    ):
        """These are the parameters that work for CIFAR-10 data.

        Args:
          resnet_size (int): The number of convolutional layers needed in the model.
          data_format (str): Either 'channels_first' or 'channels_last', specifying which
            data format to use when setting up the model.
          num_classes (int): The number of output classes needed from the model. This
            enables users to extend the same model to their own datasets.
          resnet_version (int): Integer representing which version of the ResNet network
          to use. Valid values: [1, 2]
          dtype (tensorflow.python.framework.dtypes.DType): The TensorFlow dtype to use for calculations.

        Raises:
          ValueError: if invalid resnet_size is chosen
        """
        if resnet_size % 6 != 2:
            raise ValueError("resnet_size must be 6n + 2:", resnet_size)

        num_blocks = (resnet_size - 2) // 6

        super(Cifar10Model, self).__init__(
            resnet_size=resnet_size,
            bottleneck=False,
            num_classes=num_classes,
            num_filters=16,
            kernel_size=3,
            conv_stride=1,
            first_pool_size=None,
            first_pool_stride=None,
            block_sizes=[num_blocks] * 3,
            block_strides=[1, 2, 2],
            resnet_version=resnet_version,
            data_format=data_format,
            dtype=dtype,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlbench_core.controlflow.pytorch` package."""

import pytest

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import random

from mlbench_core.controlflow.pytorch import TrainValidation
from mlbench_core.evaluation.pytorch.metrics import TopKAccuracy
from mlbench_core.lr_scheduler.pytorch import multistep_learning_rates_with_warmup
from mlbench_core.utils.pytorch.distributed import AllReduceAggregation


@pytest.fixture
def model():
    return nn.Linear(1, 2)


@pytest.fixture
def optimizer(model):
    return optim.SGD(model.parameters(), lr=0.1)


@pytest.fixture
def loss_function():
    return nn.CrossEntropyLoss()


@pytest.fixture
def metrics():
    return [TopKAccuracy(topk=1)]


@pytest.fixture
def scheduler(optimizer):
    return multistep_learning_rates_with_warmup(
        optimizer,
        1,
        0.1,
        0.1,
        [5, 10],
        warmup_duration=2,
        warmup_linear_scaling=False,
        warmup_lr=0.2,
    )


def test_instantiation(mocker, model, optimizer, loss_function, metrics, scheduler):
    mocker.patch("mlbench_core.controlflow.pytorch.controlflow.dist")

    batch_size = 2

    tv = TrainValidation(
        model,
        optimizer,
        loss_function,
        metrics,
        scheduler,
        batch_size,
        10,
        0,
        1,
        1,
        "fp32",
    )

    assert tv is not None


def test_training(mocker, model, optimizer, loss_function, metrics, scheduler):
    mocker.patch("mlbench_core.controlflow.pytorch.controlflow.dist")
    mocker.patch("mlbench_core.utils.pytorch.distributed.dist")
    mocker.patch("mlbench_core.utils.tracker.LogMetrics")

    batch_size = 2

    tv = TrainValidation(
        model,
        optimizer,
        loss_function,
        metrics,
        scheduler,
        batch_size,
        10,
        0,
        1,
        1,
        "fp32",
    )

    train_set = [random.random() for _ in range(100)]
    train_set = [
        (
            torch.FloatTensor([n * 50 - 25]),
            1 if (n > 0.5) != (random.random() < 0.1) else 0,
        )
        for n in train_set
    ]

    test_set = [random.random() for _ in range(10)]
    test_set = [
        (
            torch.FloatTensor([n * 50 - 25]),
            1 if (n > 0.5) != (random.random() < 0.1) else 0,
        )
        for n in test_set
    ]

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=False)

    tv.run(
        dataloader_train_fn=lambda: train_loader,
        dataloader_val_fn=lambda: test_loader,
        repartition_per_epoch=True,
    )

    assert tv.tracker.current_epoch == 10
    assert tv.tracker.best_epoch > -1
    assert tv.tracker.best_metric_value > 50.0



"""Evaluate training/validation set using models in checkpoints"""
import logging

from mlbench_core.utils.pytorch.distributed import AllReduceAggregation
from mlbench_core.utils.pytorch.distributed import global_average
from mlbench_core.utils.pytorch.helpers import iterate_dataloader

import torch

logger = logging.getLogger("mlbench")


class CheckpointsEvaluationControlFlow(object):
    """Evaluate models on training / validation dataset.

    Args:
        ckpt_dir (str): Path to checkpoints.
        rank (int): The rank of the current process
        world_size (int): The total number of workers
        checkpointer (:obj:`Checkpointer`): Used to load checkpoints.
        model (:obj:`torch.optim.Optimizer`): An optimizer for the given model.
        epochs (int): Number of epochs to traing.
        loss_function (:obj:`torch.nn.modules.loss._Loss`): loss function.
        metrics (:obj:`list` of :obj:`mlbench_core.evaluation.pytorch.*`): metrics like TopKAccuracy.
        use_cuda (bool): Whether to train on GPU or not. Default: `False`
        dtype (str): The datatype to use for the dataloader data
        max_batch_per_epoch (int): Maximum number of batches per epoch. Whole dataset
        is used if not specified. Default: `None`
    """

    def __init__(
        self,
        ckpt_dir,
        rank,
        world_size,
        checkpointer,
        model,
        epochs,
        loss_function,
        metrics,
        use_cuda=False,
        dtype=None,
        max_batch_per_epoch=None,
    ):
        self.ckpt_dir = ckpt_dir
        self.rank = rank
        self.checkpointer = checkpointer
        self.model = model
        self.epochs = epochs
        self.loss_function = loss_function
        self.metrics = metrics
        self.dtype = dtype
        self.max_batch_per_epoch = max_batch_per_epoch
        self.use_cuda = use_cuda

        self.model_agg_fn = AllReduceAggregation(world_size=world_size).agg_model

        self._check_checkpoints()

    def _check_checkpoints(self):
        for epoch in range(self.epochs):
            self.checkpointer.checkpoint_exists(self.ckpt_dir, self.rank, epoch)

    def _load_model(self, epoch):
        # Load epoch-rank model
        model = self.checkpointer.load_model_by_epoch(
            self.ckpt_dir, self.rank, epoch, self.model
        )

        # aggregate models
        self.model_agg_fn(model, op="avg")
        return model

    def evaluate_by_epochs(self, dataloader):
        """Evaluate dataset using the averaged models.

        In each epoch each process loads models and averages them. The averaged model is
        used to evaluate train / validation dataset.

        Args:
            dataloader (:obj:`torch.utils.data.DataLoader`): The dataset to be evaluated.

        Returns:
            list: list of stats of models in each epoch.
        """
        stats_list = []
        for epoch in range(self.epochs):
            # Same model for all workers.
            model = self._load_model(epoch)
            model.eval()

            stats = {"epoch": epoch, "count": 0, "total_loss": 0}
            for metric in self.metrics:
                stats["total_" + metric.name] = 0

            data_iter = iterate_dataloader(
                dataloader, self.dtype, self.max_batch_per_epoch, self.use_cuda
            )

            with torch.no_grad():
                for i, (data, target) in enumerate(data_iter):
                    output = model(data)

                    # Compute loss and metrics.
                    count = len(target)
                    stats["count"] += count
                    stats["total_loss"] += self.loss_function(output, target) * count
                    for metric in self.metrics:
                        stats["total_" + metric.name] += metric(output, target) * count

                    logger.info(
                        "E{:4}B{:4}: total loss={:10.3e}".format(
                            epoch, i, stats["total_loss"] / stats["count"]
                        )
                    )

            # Keep globally averaged loss / metrics, etc.
            stats["loss"] = global_average(stats["total_loss"], stats["count"]).item()
            for metric in self.metrics:
                stats[metric.name] = global_average(
                    stats["total_" + metric.name], stats["count"]
                ).item()
                del stats["total_" + metric.name]
            del stats["count"], stats["total_loss"]

            stats_list.append(stats)
        return stats_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlbench_core.utils.pytorch.helpers` package."""

import pytest
import time
import itertools
import torch

from mlbench_core.utils.pytorch.helpers import (
    Timeit,
    maybe_range,
    update_best_runtime_metric,
    convert_dtype,
    config_pytorch,
    config_path,
    iterate_dataloader,
    log_metrics,
)
from mlbench_core.utils import Tracker, LogMetrics
from mlbench_core.evaluation.pytorch.metrics import TopKAccuracy


def test_timeit():
    timeit = Timeit()

    cumu = timeit.cumu

    time.sleep(0.1)
    timeit.pause()
    new_cumu = timeit.cumu

    assert new_cumu - cumu - 0.1 < 0.01

    time.sleep(0.1)
    newer_cumu = timeit.cumu

    assert new_cumu == newer_cumu

    timeit.resume()
    time.sleep(0.1)
    timeit.pause()

    last_cumu = timeit.cumu

    assert last_cumu - newer_cumu - 0.1 < 0.01
    assert last_cumu - cumu - 0.3 < 0.01


def test_maybe_range():
    r = maybe_range(10)

    assert len(r) == 10
    assert r == range(10)

    r = maybe_range(None)

    assert isinstance(r, itertools.count)
    assert next(r) == 0
    assert next(r) == 1


def test_update_best_runtime_metric(mocker):
    tracker = Tracker([TopKAccuracy(5)], 1, 0)
    # tracker = mocker.patch('mlbench_core.utils.pytorch.helpers.Tracker')

    is_best, best_metric_name = update_best_runtime_metric(tracker, 10.0, "prec")

    assert is_best
    assert best_metric_name == "best_prec"

    is_best, best_metric_name = update_best_runtime_metric(tracker, 11.0, "prec")

    assert is_best
    assert best_metric_name == "best_prec"

    is_best, best_metric_name = update_best_runtime_metric(tracker, 9.0, "prec")

    assert not is_best
    assert best_metric_name == "best_prec"


def test_convert_dtype():
    t = torch.IntTensor([0])

    tt = convert_dtype("fp32", t)

    assert tt.dtype == torch.float32

    tt2 = convert_dtype("fp64", t)

    assert tt2.dtype == torch.float64

    with pytest.raises(NotImplementedError):
        tt3 = convert_dtype("int", t)


def test_config_pytorch(mocker):
    mocker.patch("torch.distributed.get_rank", return_value=1)
    mocker.patch("torch.distributed.get_world_size", return_value=1)
    mocker.patch("mlbench_core.utils.pytorch.helpers.FCGraph")

    rank, world_size, graph = config_pytorch(
        use_cuda=False, seed=42, cudnn_deterministic=True
    )

    assert rank == 1
    assert world_size == 1
    assert graph is not None


def test_LogMetrics(mocker):
    mocker.patch("mlbench_core.utils.pytorch.helpers.ApiClient")

    LogMetrics.log("1", 1, 1, "loss", 123)

    mocker.patch.dict("os.environ", {"MLBENCH_IN_DOCKER": "True"})

    LogMetrics.log("1", 1, 1, "loss", 123)


def test_log_metrics(mocker):
    mocker.patch("mlbench_core.utils.pytorch.helpers.ApiClient")

    log_metrics("1", 1, 1, "loss", 123)

    mocker.patch.dict("os.environ", {"MLBENCH_IN_DOCKER": "True"})

    log_metrics("1", 1, 1, "loss", 123)


def test_config_path(mocker):
    sh = mocker.patch("shutil.rmtree")
    osmk = mocker.patch("os.makedirs")

    config_path("/tmp/checkpoints", delete_existing_ckpts=False)

    osmk.assert_called_once_with("/tmp/checkpoints", exist_ok=True)
    assert sh.call_count == 0

    config_path("/tmp/checkpoints", delete_existing_ckpts=True)

    assert sh.call_count == 1
    assert osmk.call_count == 2


def test_iterate_dataloader(mocker):
    dataloader = [
        (torch.IntTensor([0]), torch.IntTensor([1])),
        (torch.IntTensor([2]), torch.IntTensor([3])),
    ]

    it = iterate_dataloader(
        dataloader, "fp32", max_batch_per_epoch=2, transform_target_type=True
    )

    first = next(it)

    assert first[0].dtype == torch.float32
    assert first[1].dtype == torch.float32
    assert first[0].data.item() == 0.0
    assert first[1].item() == 1.0

    second = next(it)

    assert second[0].dtype == torch.float32
    assert second[1].dtype == torch.float32
    assert second[0].data.item() == 2.0
    assert second[1].item() == 3.0



r"""Define tensorflow metrics."""

import tensorflow as tf


class TopKAccuracy(object):
    """Compute the top-k accuracy of logits.

    Args:
        logits (:obj:`tf.Tensor`): input tensor
        labels (:obj:`tf.Tensor`): input one-hot encoded tensor.
        topkk (:obj:`int`, optional): Defaults to 1. top k accuracy.
    """

    def __init__(self, logits, labels, topk=1):
        labels = tf.cast(labels, tf.int32)
        true_classes = tf.argmax(labels, axis=1)

        # predicted classes
        pred_probs = tf.nn.softmax(logits, name="softmax_tensor")
        pred_classes = tf.argmax(pred_probs, axis=1)

        # get metrics.
        with tf.name_scope("metrics"):
            if topk == 1:
                self.name = "Prec@1"
                self.metric_op = (
                    tf.reduce_mean(
                        tf.cast(tf.equal(true_classes, pred_classes), tf.float32)
                    )
                    * 100.0
                )
            else:
                topk_op = tf.nn.in_top_k(
                    predictions=pred_probs, targets=true_classes, k=topk
                )
                self.name = "Prec@" + str(topk)
                self.metric_op = tf.reduce_mean(tf.cast(topk_op, tf.float32)) * 100.0


def topk_accuracy_with_logits(logits, labels, k=1):
    """Compute the top-k accuracy of logits.

    Args:
        logits (:obj:`tf.Tensor`): input tensor
        labels (:obj:`tf.Tensor`): input one-hot encoded tensor.
        k (:obj:`int`, optional): Defaults to 1. top k accuracy.

    Returns:
        :obj:`tf.Tensor`: a scalar tensor of the accuracy (between 0 and 1).
    """
    m = TopKAccuracy(logits=logits, labels=labels, topk=k)
    return {"name": m.name, "value": m.metric_op}



r"""Test the tensorflow load and preprocess cifar-10 correctly.

Credit https://github.com/tensorflow/models/blob/master/tutorials/image/cifar10/cifar10_input.py
"""
import logging
import os
import sys
import tarfile
from six.moves import urllib, xrange
import tensorflow as tf


class DatasetCifar(object):
    """
    This clas is adapted from the following script
    https://github.com/tensorflow/models/blob/master/tutorials/image/cifar10/cifar10_input.py
    """

    def __init__(
        self,
        dataset,
        dataset_root,
        batch_size,
        world_size,
        rank,
        seed,
        tf_dtype=tf.float32,
    ):
        """
        Args:
            dataset (str): Name of the dataset e.g. `cifar-10`, `cifar-100`.
            dataset_root (str): Root directory to the dataset.
            batch_size (int): Size of batch.
            world_size (int): Size of the world size.
            rank (int): Rank of the process.
            seed (int): Seed of random number.
            tf_dtype (tensorflow.python.framework.dtypes.DType, optional): Defaults to tf.float32.
                Datatypes of the tensor.
        """

        # define image size and some commonly used parameters.
        self.data_url = "http://www.cs.toronto.edu/~kriz/{}-binary.tar.gz".format(
            dataset
        )

        self.dataset = dataset
        self.dataset_dir = dataset_root
        self.seed = seed

        self.batch_size = batch_size * world_size
        self.num_examples_per_epoch_for_train = 50000
        self.num_examples_per_epoch_for_eval = 10000
        self.num_batches_per_epoch_for_train = (
            self.num_examples_per_epoch_for_train // self.batch_size
        )
        self.num_batches_per_epoch_for_eval = (
            self.num_examples_per_epoch_for_eval // self.batch_size
        )

        self.image_size = 32
        self.image_channel = 3

        if dataset == "cifar-10":
            self.label_bytes = 1
            self.label_offset = 0
            self.num_classes = 10
        else:
            self.label_bytes = 1
            self.label_offset = 1
            self.num_classes = 100

        # Every record consists of a label followed by the image,
        # with a fixed number of bytes for each.
        self.image_bytes = self.image_size * self.image_size * self.image_channel
        self.record_bytes = self.label_bytes + self.label_offset + self.image_bytes

        # download the dataset.
        self.maybe_download_and_extract()

        # Define dataset for both training and validation
        self.train_dataset = self.input_fn(
            is_train=True, repeat_count=-1, num_shards=world_size, shard_index=rank
        )
        self.validation_dataset = self.input_fn(
            is_train=False, repeat_count=-1, num_shards=world_size, shard_index=rank
        )

        # Define reinitializable iterators for dataset.
        # Initialize operations like train_init_op and validation_init_op when switch mode.
        iterator = tf.data.Iterator.from_structure(
            self.train_dataset.output_types, self.train_dataset.output_shapes
        )
        self.train_init_op = iterator.make_initializer(self.train_dataset)
        self.validation_init_op = iterator.make_initializer(self.validation_dataset)

        # do not need to feed_dict for `inputs`, `labels`, `training`
        (self.inputs, self.labels), self.training = iterator.get_next()

    def maybe_download_and_extract(self):
        """Download and extract the tarball from Alex's website."""
        dest_directory = os.path.join(self.dataset_dir, self.dataset)

        if not os.path.exists(dest_directory):
            os.makedirs(dest_directory)
        filename = self.data_url.split("/")[-1]
        filepath = os.path.join(dest_directory, filename)

        if not os.path.exists(filepath):

            def _progress(count, block_size, total_size):
                sys.stdout.write(
                    "\r>> Downloading %s %.1f%%"
                    % (filename, float(count * block_size) / float(total_size) * 100.0)
                )
                sys.stdout.flush()

            filepath, _ = urllib.request.urlretrieve(self.data_url, filepath, _progress)
            logging.debug("download file to the path:" + filepath)
        else:
            logging.debug("retrieve file to the path:" + filepath)

        if self.dataset == "cifar-10":
            self.data_dir = os.path.join(dest_directory, self.dataset + "-batches-bin")
        else:
            self.data_dir = os.path.join(dest_directory, self.dataset + "-binary")

        if not os.path.exists(self.data_dir):
            logging.debug("does not exist extracted file: {}".format(self.data_dir))
            tarfile.open(filepath, "r:gz").extractall(dest_directory)

    def record_dataset(self, filenames):
        """Returns an input pipeline Dataset from `filenames`."""
        return tf.data.FixedLengthRecordDataset(filenames, self.record_bytes)

    def get_filenames(self, is_training=True):
        if is_training:
            filenames = (
                [
                    os.path.join(self.data_dir, "data_batch_%d.bin" % i)
                    for i in xrange(1, 6)
                ]
                if self.dataset == "cifar-10"
                else [os.path.join(self.data_dir, "train.bin")]
            )
        else:
            filenames = (
                [os.path.join(self.data_dir, "test_batch.bin")]
                if self.dataset == "cifar-10"
                else [os.path.join(self.data_dir, "test.bin")]
            )
        return filenames

    def parse_record(self, raw_record):
        """Parse CIFAR-10/100 image and label from a raw record."""
        # Convert bytes to a vector of uint8 that is record_bytes long.
        # record_vector = tf.decode_raw(raw_record, tf.uint8)
        record = tf.reshape(tf.decode_raw(raw_record, tf.uint8), [self.record_bytes])

        # The first byte represents the label,
        # which we convert from uint8 to int32 and then to one-hot.
        label = tf.squeeze(
            tf.cast(tf.slice(record, [self.label_offset], [self.label_bytes]), tf.int32)
        )
        label = tf.one_hot(label, self.num_classes)

        # The remaining bytes after the label represent the image,
        # which we reshape from
        # [depth * height * width] to [depth, height, width].
        depth_major = tf.reshape(
            tf.slice(
                record, [self.label_offset + self.label_bytes], [self.image_bytes]
            ),
            [self.image_channel, self.image_size, self.image_size],
        )

        # Convert from [depth, height, width] to [height, width, depth],
        # and cast as float32.
        image = tf.cast(tf.transpose(depth_major, [1, 2, 0]), tf.float32)
        return image, label

    def preprocess_image(self, image, is_training):
        """Preprocess a single image of layout [height, width, depth]."""
        if is_training:
            # Randomly flip the image horizontally.
            image = tf.image.random_flip_left_right(image)

            # decide to augment the data and only for training data.
            # Resize the image to add four extra pixels on each side.
            pad = 4
            image = tf.image.resize_image_with_crop_or_pad(
                image, self.image_size + pad * 2, self.image_size + pad * 2
            )

            # Randomly crop a [image_size, image_size] section of the image.
            image = tf.random_crop(
                image, [self.image_size, self.image_size, self.image_channel]
            )

            # consider to add more stronger augmentation.
            # image = tf.image.random_brightness(image, max_delta=63)
            # image = tf.image.random_contrast(image, lower=0.2, upper=1.8)
        else:
            image = tf.image.resize_image_with_crop_or_pad(
                image, self.image_size, self.image_size
            )

        if self.dataset == "cifar-10":
            stats = {
                "mean": tf.constant([0.4914, 0.4822, 0.4465], dtype=tf.float32),
                "std": tf.constant([0.2023, 0.1994, 0.2010], dtype=tf.float32),
            }
            image = (image / 256 - stats["mean"]) / stats["std"]
        else:
            # Subtract off the mean and divide by the variance of the pixels.
            # image = tf.image.per_image_standardization(image)
            raise NotImplementedError
        return image

    def input_fn(self, is_train, repeat_count=-1, num_shards=1, shard_index=0):
        """Input_fn using the tf.data input pipeline for CIFAR-10 dataset.

        In synchronized training, faster nodes may use more batches than the number
        of batches availble. Thus repeat dataset for engouh times to avoid throwing
        error.

        In the distributed settings, datasets are split into `num_shards`
        non-overlapping parts and each process takes one shard by its index.

        Args:
            is_train (bool): A boolean denoting whether the input is for training.
            repeat_count (int): Defaults to -1. Count of dataset repeated times
                with -1 for infinite.
            num_shards (int): Defaults to 1. Number of Shards the dataset is splitted.
            shard_index (int): Defaults to 0. Index of shard to use.

        Returns
            tf.data.Dataset object of `((inputs, labels), is_train)`.
        """

        dataset = self.record_dataset(self.get_filenames(is_train))

        dataset = dataset.shard(num_shards, shard_index)

        if is_train:
            # When choosing shuffle buffer sizes, larger sizes result in better
            # randomness, while smaller sizes have better performance.
            # Because CIFAR-10 is a relatively small dataset,
            # we choose to shuffle the full epoch.
            dataset = dataset.shuffle(
                buffer_size=self.num_examples_per_epoch_for_train,
                seed=self.seed,
                reshuffle_each_iteration=True,
            )

        dataset = dataset.map(self.parse_record, num_parallel_calls=8)
        dataset = dataset.map(
            lambda image, label: (self.preprocess_image(image, is_train), label),
            num_parallel_calls=8,
        )
        # TODO: change prefetch size? to `tf.contrib.data.AUTOTUNE`
        # https://github.com/tensorflow/models/blob/master/official/resnet/resnet_run_loop.py#L103
        dataset = dataset.prefetch(2 * self.batch_size)

        # We call repeat after shuffling, rather than before,
        # to prevent separate epochs from blending together.
        dataset = dataset.repeat(repeat_count)

        # Batch results by up to batch_size,
        # and then fetch the tuple from the iterator.
        dataset = dataset.batch(self.batch_size)

        # A boolean indicating the mode / dataset type : True for training and False for validation.
        mode = tf.data.Dataset.from_tensor_slices([is_train]).repeat()
        return tf.data.Dataset.zip((dataset, mode))



"""Customized Loss Functions."""

import torch.nn.functional as F
from torch.nn.modules.loss import _WeightedLoss
from torch import nn
import torch


class BCELossRegularized(_WeightedLoss):
    """Binary Cross Entropy (BCE) with l1/l2 regularization.

    Args:
        weight (Tensor, optional): a manual rescaling weight given to each
            class. If given, it has to be a Tensor of size `C`. Otherwise,
            it is
            treated as if having all ones.
        size_average (bool, optional): Deprecated (see :attr:`reduction`).
        By default,
            the losses are averaged over each loss element in the batch.
            Note that for
            some losses, there multiple elements per sample. If the field
            :attr:`size_average`
            is set to ``False``, the losses are instead summed for each
            minibatch. Ignored
            when reduce is ``False``. Default: ``True``
        reduce (bool, optional): Deprecated (see :attr:`reduction`). By
        default, the
            losses are averaged or summed over observations for each
            minibatch depending
            on :attr:`size_average`. When :attr:`reduce` is ``False``,
            returns a loss per
            batch element instead and ignores :attr:`size_average`. Default:
            ``True``
        l1 (float, optional): The scale of the L1 regularization. Default:
        ``0.0``
        l2 (float, optional): The scale of the L2 regularization. Default:
        ``0.0``
        model (:obj:`torch.nn.Module`): a pytorch model to be trained and
        validated.
        reduction (string, optional): Specifies the reduction to apply to
        the output:
            'none' | 'elementwise_mean' | 'sum'. 'none': no reduction will
            be applied,
            'elementwise_mean': the sum of the output will be divided by the
            number of
            elements in the output, 'sum': the output will be summed. Note:
            :attr:`size_average`
            and :attr:`reduce` are in the process of being deprecated,
            and in the meantime,
            specifying either of those two args will override
            :attr:`reduction`. Default: 'elementwise_mean'
    """

    def __init__(
        self,
        weight=None,
        size_average=None,
        reduce=None,
        l1=0.0,
        l2=0.0,
        model=None,
        reduction="elementwise_mean",
    ):
        super(BCELossRegularized, self).__init__(
            weight, size_average, reduce, reduction
        )
        self.l2 = l2
        self.l1 = l1
        self.model = model

    def forward(self, input_, target):
        output = F.binary_cross_entropy(
            input_, target.float(), weight=self.weight, reduction=self.reduction
        )
        l2_loss = sum(param.norm(2) ** 2 for param in self.model.parameters())
        output += self.l2 / 2 * l2_loss
        l1_loss = sum(param.norm(1) for param in self.model.parameters())
        output += self.l1 * l1_loss
        return output


class MSELossRegularized(_WeightedLoss):
    """Mean Squared Error (MSE) with l1/l2 regularization.

    Args:
        weight (Tensor, optional): a manual rescaling weight given to each
            class. If given, it has to be a Tensor of size `C`. Otherwise,
            it is
            treated as if having all ones.
        size_average (bool, optional): Deprecated (see :attr:`reduction`).
        By default,
            the losses are averaged over each loss element in the batch.
            Note that for
            some losses, there multiple elements per sample. If the field
            :attr:`size_average`
            is set to ``False``, the losses are instead summed for each
            minibatch. Ignored
            when reduce is ``False``. Default: ``True``
        reduce (bool, optional): Deprecated (see :attr:`reduction`). By
        default, the
            losses are averaged or summed over observations for each
            minibatch depending
            on :attr:`size_average`. When :attr:`reduce` is ``False``,
            returns a loss per
            batch element instead and ignores :attr:`size_average`. Default:
            ``True``
        l1 (float, optional): The scale of the L1 regularization. Default:
        ``0.0``
        l2 (float, optional): The scale of the L2 regularization. Default:
        ``0.0``
        model (:obj:`torch.nn.Module`): a pytorch model to be trained and
        validated.
        reduction (string, optional): Specifies the reduction to apply to
        the output:
            'none' | 'elementwise_mean' | 'sum'. 'none': no reduction will
            be applied,
            'elementwise_mean': the sum of the output will be divided by the
            number of
            elements in the output, 'sum': the output will be summed. Note:
            :attr:`size_average`
            and :attr:`reduce` are in the process of being deprecated,
            and in the meantime,
            specifying either of those two args will override
            :attr:`reduction`. Default: 'elementwise_mean'
    """

    def __init__(
        self,
        weight=None,
        size_average=None,
        reduce=None,
        l1=0.0,
        l2=0.0,
        model=None,
        reduction="elementwise_mean",
    ):
        super(MSELossRegularized, self).__init__(
            weight, size_average, reduce, reduction
        )
        self.l2 = l2
        self.l1 = l1
        self.model = model

    def forward(self, input_, target):
        output = F.mse_loss(input_, target, reduction=self.reduction)
        l2_loss = sum(param.norm(2) ** 2 for param in self.model.parameters())
        output += self.l2 / 2 * l2_loss
        l1_loss = sum(param.norm(1) for param in self.model.parameters())
        output += self.l1 * l1_loss
        return output


class LabelSmoothing(nn.Module):
    """
    NLL loss with label smoothing.
    """

    def __init__(self, padding_idx, smoothing=0.0):
        """
        Constructor for the LabelSmoothing module.

        Args:
            padding_idx (int): Code for padding char
            smoothing (float): Smoothing value
        """
        super(LabelSmoothing, self).__init__()
        self.padding_idx = padding_idx
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing

    def forward(self, x, target):
        logprobs = torch.nn.functional.log_softmax(x, dim=-1).type(torch.float32)

        non_pad_mask = target != self.padding_idx
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)[non_pad_mask]
        smooth_loss = -logprobs.mean(dim=-1)[non_pad_mask]
        loss = self.confidence * nll_loss + self.smoothing * smooth_loss
        return loss.sum()



# -*- coding: utf-8 -*-

import math
from bisect import bisect_right

import numpy as np
from torch.optim.lr_scheduler import LambdaLR


def const(optimizer):
    return LambdaLR(optimizer, lr_lambda=lambda x: 1.0)


def triangular_learning_rates(
    optimizer, base_lr, max_lr, cycle_length, scale_fn, extra, mode
):
    """ Linearily Scale Learning Rate

    If one cycle is applied with length smaller than the total number of iterations, then
    use small learning rate for the remaining iterations.

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        base_lr (float): Lower bound and initial learning rate in a cycle.
        max_lr (float): Upper bound in a cycle
        cycle_length (int): Length of a cycle in terms of batches.
        scale_fn(:func:`Function`): The scaling function.
        extra (int): The number of extra epochs to perform after a cycle
        mode (str): The scaling mode to use. One of `linear`, `triangular`, `one_cycle`,
            `triangular2` or `exp_range`

    Returns:
        A learning rate scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`)
    """
    step_size = cycle_length / 2

    if mode == "one_cycle":

        def f(iterations):
            if iterations <= cycle_length:
                cycle = np.floor(1 + iterations / (2 * step_size))
                x = np.abs(iterations / step_size - 2 * cycle + 1)
                lr = base_lr + (max_lr - base_lr) * np.maximum(0, (1 - x)) * scale_fn(
                    cycle, iterations
                )
            else:
                lr = base_lr * extra
            return lr / base_lr

    else:

        def f(iterations):
            cycle = np.floor(1 + iterations / (2 * step_size))
            x = np.abs(iterations / step_size - 2 * cycle + 1)
            lr = base_lr + (max_lr - base_lr) * np.maximum(0, (1 - x)) * scale_fn(
                cycle, iterations
            )
            return lr / base_lr

    # Use base_lr to overwrite the --lr
    for group in optimizer.param_groups:
        group["initial_lr"] = base_lr
    optimizer.base_lrs = [base_lr for _ in optimizer.param_groups]
    return LambdaLR(optimizer, lr_lambda=f)


def cyclical_learning_rates(
    optimizer, mode, gamma, cycle_length, base_lr, max_lr, extra_epochs
):
    """ Cyclically Scale Learning Rate

    If one cycle is applied with length smaller than the total number of iterations, then
    use small learning rate for the remaining iterations.

    Since :cite:`smith2017cyclical` mentioned that triangular, Welch, Hann windows produce equivalent results,
    we only implement triangular learning rate policy, also known as **linear cycle**.

    The original implementation of :cite:`smith2017cyclical` can be found from
    `here <https://github.com/bckenstler/CLR>`_.

    :cite:`smith2017super` uses one cycle with extra epochs.

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        mode (str): The scaling mode to use. One of `linear`, `triangular`, `one_cycle`,
            `triangular2` or `exp_range`
        base_lr (float): Lower bound and initial learning rate in a cycle.
        max_lr (float): Upper bound in a cycle
        max_lr (float): The maximum learning rate
        extra_epochs (int): The number of extra epochs to perform after a cycle

    Returns:
        A learning rate scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`)
    """
    if mode in ["linear", "triangular", "one_cycle"]:

        def scale_fn(cycle, iterations):
            return 1.0

    elif mode == "triangular2":

        def scale_fn(cycle, iterations):
            return 1 / (2.0 ** (cycle - 1))

    elif mode == "exp_range":

        def scale_fn(cycle, iterations):
            return gamma ** iterations

    else:
        raise ValueError("Cycle mode {} not support.".format(mode))

    return triangular_learning_rates(
        optimizer,
        base_lr,
        max_lr,
        cycle_length=cycle_length,
        scale_fn=scale_fn,
        extra=extra_epochs,
        mode=mode,
    )


def multistep_learning_rates_with_warmup(
    optimizer,
    world_size,
    lr,
    gamma,
    milestones,
    warmup_duration=None,
    warmup_lr=None,
    warmup_linear_scaling=False,
):
    """ Multistep Learning Rate Schedule with warmup

    In :cite:`goyal2017accurate`, warmup is used in order to apply the ``Linear Scaling Rule``.
    Starting from the ``base_lr``, lr gradually increases to ``base_lr * scaling_factor``.
    Then use multiply the learning rate by ``gamma`` at specified milestones.
    See :cite:`ginsburg2018large`

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        world_size (int): The total number of workers
        lr (float): The initial learning rate
        gamma (float): Decay factor for learning rate
        milestones (:obj:`list` of :obj:`int`): The epochs/steps at which to reduce the
            learning rate
        warmup_duration (int): The number of epochs to perform warmup before regular
            lr scaling starts. Default: `None`
        warmup_lr (float): The learning rate to use for the warmup epochs. Default: `None`
        warmup_linear_scaling (bool): Whether or not to linearily scale lr during
            warmup. Default: `False`
    Returns:
        A learning rate scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`)
    """
    if bool(warmup_duration) != bool(warmup_lr):
        raise ValueError(
            "Either both or none of warmup_duration and warmup_lr have to be set"
        )

    scaling_factor = 1

    if warmup_linear_scaling:
        scaling_factor = world_size

    base_lr = lr * scaling_factor

    warmup_init_lr = lr
    if warmup_lr:
        warmup_init_lr = warmup_lr

    if list(milestones) != sorted(milestones):
        raise ValueError(
            "Milestones should be a list of increasing integers."
            "Got {}".format(milestones)
        )

    if warmup_duration >= milestones[0]:
        raise ValueError(
            "The scaling phase should be earlier than the first milestone."
            "Got {} and {}".format(warmup_duration, milestones[0])
        )

    def f(duration):
        if warmup_lr and duration <= warmup_duration:
            warmup_progress = duration / warmup_duration
            lr = warmup_progress * base_lr + (1 - warmup_progress) * warmup_init_lr
        else:
            lr = base_lr * gamma ** bisect_right(milestones, duration)
        return lr / base_lr

    for group in optimizer.param_groups:
        group["initial_lr"] = base_lr
    optimizer.base_lrs = [base_lr for _ in optimizer.param_groups]
    return LambdaLR(optimizer, lr_lambda=f)


class MultistepLearningRatesWithWarmup(LambdaLR):
    """ Multistep Learningrate Scheduler with Warmup Period

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        world_size (int): The total number of workers
        gamma (float): Decay factor for learning rate
        milestones (:obj:`list` of :obj:`int`): The epochs/steps at which to reduce the
            learning rate
        lr (float): The initial learning rate
        warmup_duration (int): The number of epochs to perform warmup before regular
            lr scaling starts. Default: `None`
        warmup_linear_scaling (bool): Whether or not to linearily scale lr during
            warmup. Default: `False`
        warmup_init_lr (float): The initial learning rate to use for the warmup epochs. Default: `None`
    Returns:
        A learning rate scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`)
    """

    def __init__(
        self,
        optimizer,
        world_size,
        gamma,
        milestones,
        lr,
        warmup_duration,
        warmup_linear_scaling=True,
        warmup_init_lr=None,
    ):
        if list(milestones) != sorted(milestones):
            raise ValueError(
                "Milestones should be a list of increasing integers."
                "Got {}".format(milestones)
            )

        if warmup_duration >= milestones[0]:
            raise ValueError(
                "The scaling phase should be earlier than the first milestone."
                "Got {} and {}".format(warmup_duration, milestones[0])
            )

        self.optimizer = optimizer
        self.gamma = gamma
        self.milestones = milestones
        self.warmup_duration = warmup_duration
        self.warmup_init_lr = warmup_init_lr or lr

        scaling_factor = world_size if warmup_linear_scaling else 1
        self.warmup_scaled_lr = scaling_factor * lr

        # overwrite initial lr
        self.base_lr = lr
        for group in self.optimizer.param_groups:
            group["initial_lr"] = self.base_lr

        super(MultistepLearningRatesWithWarmup, self).__init__(self.optimizer, self.f)

    def f(self, duration):
        # warmup_lr => lr or lr * world_size => ....
        if duration <= self.warmup_duration:
            progress = duration / self.warmup_duration
            lr = progress * self.warmup_scaled_lr + (1 - progress) * self.warmup_init_lr
        else:
            lr = self.warmup_scaled_lr * self.gamma ** bisect_right(
                self.milestones, duration
            )
        return lr / self.base_lr


class SparsifiedSGDLR(LambdaLR):
    """ Learning rate schedule for sparsifiedSGD (gamma / l2_coef * (t + shifting_param))

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        gamma (float): The constant value in the numerator of the learning rate schedule formula
        l2_coef (float): The regularization rate which is used in the denominator of the learning rate schedule formula
        shifting_param (float): The constant value in the denominator of the learning rate schedule formula
    """

    def __init__(self, optimizer, gamma, l2_coef, shifting_param):
        self.shifting_param = shifting_param
        self.optimizer = optimizer

        for group in self.optimizer.param_groups:
            group["initial_lr"] = gamma / l2_coef

        self.optimizer.base_lrs = [gamma / l2_coef for _ in self.optimizer.param_groups]

        super(SparsifiedSGDLR, self).__init__(self.optimizer, self.f)

    def f(self, iteration):
        return 1 / max(1, (self.shifting_param + iteration))


class TimeDecayLR(LambdaLR):
    """
    Time based decay learning rate schedule for SGD (alpha / (t + beta))

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        alpha (float): The constant value in the numerator of the learning rate schedule formula
        beta (float): The constant value in the denominator of the learning rate schedule formula
    Returns:
        A learning rate scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`)
    """

    def __init__(self, optimizer, alpha, beta):
        self.alpha = alpha
        self.beta = beta
        self.optimizer = optimizer

        for group in self.optimizer.param_groups:
            group["initial_lr"] = alpha / beta

        self.optimizer.base_lrs = [alpha / beta for _ in self.optimizer.param_groups]

        super(TimeDecayLR, self).__init__(self.optimizer, self.f)

    def f(self, iteration):
        return self.beta / (self.beta + iteration)


class SQRTTimeDecayLR(LambdaLR):
    """
    Time based decay learning rate schedule for SGD (alpha / sqrt(t))

    Args:
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        alpha (float): The constant value in the numerator of the learning rate schedule formula
    Returns:
        A learning rate scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`)
    """

    def __init__(self, optimizer, alpha):
        self.alpha = alpha
        self.optimizer = optimizer

        for group in self.optimizer.param_groups:
            group["initial_lr"] = alpha

        self.optimizer.base_lrs = [alpha for _ in self.optimizer.param_groups]

        super(SQRTTimeDecayLR, self).__init__(self.optimizer, self.f)

    def f(self, iteration):
        return 1.0 / math.sqrt(max(1, iteration))


def perhaps_convert_float(param, total):
    if isinstance(param, float):
        param = int(param * total)
    return param


class ExponentialWarmupMultiStepLR(LambdaLR):
    """
    Learning rate scheduler with exponential warmup and step decay.
    """

    def __init__(
        self,
        optimizer,
        iterations,
        warmup_steps=0,
        remain_steps=1.0,
        decay_interval=None,
        decay_steps=4,
        decay_factor=0.5,
    ):
        """
        Constructor of Exponential WarmupMultiStepLR.

        Parameters: warmup_steps, remain_steps and decay_interval accept both
        integers and floats as an input. Integer input is interpreted as
        absolute index of iteration, float input is interpreted as a fraction
        of total training iterations (epochs * steps_per_epoch).

        If decay_interval is None then the decay will happen at regulary spaced
        intervals ('decay_steps' decays between iteration indices
        'remain_steps' and 'iterations').

        Args:
            optimizer: instance of optimizer
            iterations (int): total number of training iterations
            warmup_steps (int): number of warmup iterations
            remain_steps (int|float): start decay at 'remain_steps' iteration
            decay_interval (int|float): interval between LR decay steps
            decay_steps (int): max number of decay steps
            decay_factor (float): decay factor
        """

        # iterations before learning rate reaches base LR
        self.warmup_steps = perhaps_convert_float(warmup_steps, iterations)

        # iteration at which decay starts
        self.remain_steps = perhaps_convert_float(remain_steps, iterations)

        # number of steps between each decay
        if decay_interval is None:
            # decay at regulary spaced intervals
            decay_iterations = iterations - self.remain_steps
            self.decay_interval = decay_iterations // decay_steps
            self.decay_interval = max(self.decay_interval, 1)
        else:
            self.decay_interval = perhaps_convert_float(decay_interval, iterations)

        # multiplicative decay factor
        self.decay_factor = decay_factor

        # max number of decay steps
        self.decay_steps = decay_steps

        if self.warmup_steps > self.remain_steps:
            self.warmup_steps = self.remain_steps

        super(ExponentialWarmupMultiStepLR, self).__init__(optimizer, self.f)

    def f(self, duration):
        factor = 1
        if duration <= self.warmup_steps:
            # exponential lr warmup
            if self.warmup_steps != 0:
                warmup_factor = math.exp(math.log(0.01) / self.warmup_steps)
            else:
                warmup_factor = 1.0
            factor = warmup_factor ** (self.warmup_steps - self.last_epoch)

        elif self.last_epoch >= self.remain_steps:
            # step decay
            decay_iter = self.last_epoch - self.remain_steps
            num_decay_steps = decay_iter // self.decay_interval + 1
            num_decay_steps = min(num_decay_steps, self.decay_steps)
            factor = self.decay_factor ** num_decay_steps
        return factor



import os

from mlbench_core.dataset.translation.pytorch import config
from mlbench_core.dataset.translation.pytorch.tokenizer import WMT14Tokenizer
from torchtext.data import Example, Dataset
from mlbench_core.dataset.util.tools import maybe_download_and_extract_tar_gz


def _construct_filter_pred(min_len, max_len):
    """
    Constructs a filter predicate
    Args:
        min_len (int): Min sentence length
        max_len (int): Max sentence length

    Returns:
        func
    """
    filter_pred = lambda x: not (x[0] < min_len or x[1] < min_len)
    if max_len is not None:
        filter_pred = lambda x: not (
            x[0] < min_len or x[0] > max_len or x[1] < min_len or x[1] > max_len
        )

    return filter_pred


def process_data(path, filter_pred, fields, lazy=False, max_size=None):
    """
    Loads data from the input file.
    """

    src_path, trg_path = tuple(os.path.expanduser(path + x) for x in config.EXTS)
    examples = []
    with open(src_path, mode="r", encoding="utf-8") as src_file, open(
        trg_path, mode="r", encoding="utf-8"
    ) as trg_file:
        for src_line, trg_line in zip(src_file, trg_file):
            src_line, trg_line = src_line.strip(), trg_line.strip()

            should_consider = filter_pred(
                (src_line.count(" ") + 1, trg_line.count(" ") + 1)
            )
            if src_line != "" and trg_line != "" and should_consider:
                if lazy:
                    examples.append((src_line, trg_line))
                else:
                    examples.append(Example.fromlist([src_line, trg_line], fields))

            if max_size and len(examples) >= max_size:
                break
    return examples


class WMT14Dataset(Dataset):
    urls = [
        (
            "https://storage.googleapis.com/mlbench-datasets/translation/wmt16_en_de.tar.gz",
            "wmt16_en_de.tar.gz",
        )
    ]
    name = "wmt14"
    dirname = ""

    def __init__(
        self,
        root,
        batch_first=False,
        include_lengths=False,
        lang=None,
        math_precision=None,
        download=True,
        train=False,
        validation=False,
        lazy=False,
        min_len=0,
        max_len=None,
        max_size=None,
    ):
        """WMT14 Dataset.

        Loads WMT14 dataset.
        Based on `torchtext.datasets.WMT14`

        Args:
            root (str): Root folder of WMT14 dataset
            download (bool): Download dataset
            train (bool): Whether to get the train or validation set.
                Default=True
            batch_first (bool): if True the model uses (batch,seq,feature)
                tensors, if false the model uses (seq, batch, feature)
        """

        self.lazy = lazy

        super(WMT14Dataset, self).__init__(examples=[], fields={})
        if download:
            url, file_name = self.urls[0]
            maybe_download_and_extract_tar_gz(root, file_name, url)

        src_tokenizer = WMT14Tokenizer(
            root,
            batch_first=batch_first,
            include_lengths=include_lengths,
            lang=lang,
            math_precision=math_precision,
        )
        trg_tokenizer = WMT14Tokenizer(
            root,
            batch_first=batch_first,
            include_lengths=include_lengths,
            lang=lang,
            math_precision=math_precision,
            is_target=True,
        )

        self.vocab_size = src_tokenizer.vocab_size
        self.list_fields = [("src", src_tokenizer), ("trg", trg_tokenizer)]

        self.fields = dict(self.list_fields)
        self.max_len = max_len
        self.min_len = min_len

        if train:
            path = os.path.join(root, config.TRAIN_FNAME)
        elif validation:
            path = os.path.join(root, config.VAL_FNAME)
        else:
            raise NotImplementedError()

        self.examples = process_data(
            path,
            filter_pred=_construct_filter_pred(min_len, max_len),
            fields=self.list_fields,
            lazy=lazy,
            max_size=max_size,
        )

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, item):
        if self.lazy:
            src_line, trg_line = self.examples[item]
            return Example.fromlist([src_line, trg_line], self.list_fields)
        else:
            return self.examples[item]

    def __iter__(self):
        for x in self.examples:
            if self.lazy:
                src_line, trg_line = x
                yield Example.fromlist([src_line, trg_line], self.list_fields)
            else:
                yield x



import socket
import torch
import torch.distributed as dist
from mlbench_core.utils.pytorch.distributed import get_backend_tensor


def _ranks_on_same_node(rank, world_size):
    hostname = socket.gethostname()
    hostname_length = get_backend_tensor(torch.IntTensor([len(hostname)]))

    dist.all_reduce(hostname_length, op=dist.ReduceOp.MAX)
    max_hostname_length = hostname_length.item()

    encoding = [ord(c) for c in hostname]
    encoding += [-1 for c in range(max_hostname_length - len(hostname))]
    encoding = get_backend_tensor(torch.IntTensor(encoding))

    all_encodings = [
        get_backend_tensor(torch.IntTensor([0] * max_hostname_length))
        for _ in range(world_size)
    ]
    dist.all_gather(all_encodings, encoding)

    if dist.get_backend() == dist.Backend.NCCL:
        all_encodings = [ec.cpu() for ec in all_encodings]

    all_encodings = [ec.numpy().tolist() for ec in all_encodings]

    ranks = []
    for i in range(world_size):
        if all_encodings[rank] == all_encodings[i]:
            ranks.append(i)
    return ranks


class FCGraph(object):
    """Fully-Connected Network Graph

    Args:
        config (dict): a global object containing all of the config.
    """

    def __init__(self, rank, world_size, use_cuda=False):
        self.rank = rank
        self.world_size = world_size
        self.use_cuda = use_cuda

    @property
    def current_device_name(self):
        return "cuda:{}".format(torch.cuda.current_device()) if self.use_cuda else "cpu"

    @property
    def current_device(self):
        return torch.device(self.current_device_name())

    def assigned_gpu_id(self):
        num_gpus_on_device = torch.cuda.device_count()
        ranks = _ranks_on_same_node(self.rank, self.world_size)
        # raise NotImplementedError(self.rank, ranks)
        assigned_id = ranks.index(self.rank) % num_gpus_on_device
        torch.cuda.set_device(assigned_id)

    def __str__(self):
        return "{}".format(self.current_device_name)

    def __repr__(self):
        return self.__str__()



import torch
from mlbench_core.dataset.translation.pytorch.config import BOS, EOS
from mlbench_core.evaluation.pytorch.inference.beam_search import SequenceGenerator


class Translator:
    """
    Translator that output translated sentences from GNMT model by using a Sequence Generator
    """

    def __init__(
        self,
        model,
        trg_tokenizer,
        beam_size=5,
        len_norm_factor=0.6,
        len_norm_const=5.0,
        cov_penalty_factor=0.1,
        max_seq_len=150,
    ):

        self.model = model
        self.tokenizer = trg_tokenizer
        self.insert_target_start = [BOS]
        self.insert_src_start = [BOS]
        self.insert_src_end = [EOS]
        self.batch_first = model.batch_first
        self.beam_size = beam_size

        self.generator = SequenceGenerator(
            model=self.model,
            beam_size=beam_size,
            max_seq_len=max_seq_len,
            len_norm_factor=len_norm_factor,
            len_norm_const=len_norm_const,
            cov_penalty_factor=cov_penalty_factor,
        )

    def get_detokenized_target(self, trg, batch_size):
        targets = []
        if self.batch_first:
            for i in range(batch_size):
                t = trg[i]
                targets.append(self.tokenizer.detokenize(t.tolist()))
        else:
            for i in range(batch_size):
                t = trg[:, i]
                targets.append(self.tokenizer.detokenize(t.tolist()))

        return targets

    def translate(self, src, trg):
        """ Given a source a target tokenized tensors, outputs the
        non-tokenized translation from the model, as well as the non-tokenized target

        Args:
            src:
            trg:

        Returns:

        """
        src, src_len = src
        trg, trg_len = trg
        device = next(self.model.parameters()).device

        if self.batch_first:
            batch_size = src.shape[0]
        else:
            batch_size = src.shape[1]

        bos = [self.insert_target_start] * (batch_size * self.beam_size)
        bos = torch.tensor(bos, dtype=torch.int64, device=device)
        if self.batch_first:
            bos = bos.view(-1, 1)
        else:
            bos = bos.view(1, -1)

        if self.beam_size == 1:
            generator = self.generator.greedy_search
        else:
            generator = self.generator.beam_search

        with torch.no_grad():
            context = self.model.encode(src, src_len)
            context = [context, src_len, None]
            preds, lengths, counter = generator(batch_size, bos, context)

        preds = preds.cpu()
        targets = self.get_detokenized_target(trg, batch_size)

        output = []
        for pred in preds:
            pred = pred.tolist()
            detok = self.tokenizer.detokenize(pred)
            output.append(detok)

        return output, targets



import itertools

import mlbench_core.dataset.translation.pytorch.config as config
import torch
import torch.nn as nn
from mlbench_core.models.pytorch.gnmt import BahdanauAttention
from mlbench_core.models.pytorch.gnmt.utils import init_lstm_


class RecurrentAttention(nn.Module):
    """
    LSTM wrapped with an attention module.
    """

    def __init__(
        self,
        input_size=1024,
        context_size=1024,
        hidden_size=1024,
        num_layers=1,
        batch_first=False,
        dropout=0.2,
        init_weight=0.1,
    ):
        """
        Constructor for the RecurrentAttention.

        Args:
            input_size (int): number of features in input tensor
            context_size (int): number of features in output from encoder
            hidden_size (int): internal hidden size
            num_layers (int): number of layers in LSTM
            batch_first (bool): if True the model uses (batch,seq,feature)
                tensors, if false the model uses (seq, batch, feature)
            dropout (float): probability of dropout (on input to LSTM layer)
            init_weight (float): range for the uniform initializer
        """

        super(RecurrentAttention, self).__init__()

        self.rnn = nn.LSTM(
            input_size, hidden_size, num_layers, bias=True, batch_first=batch_first
        )
        init_lstm_(self.rnn, init_weight)

        self.attn = BahdanauAttention(
            hidden_size,
            context_size,
            context_size,
            normalize=True,
            batch_first=batch_first,
        )

        self.dropout = nn.Dropout(dropout)

    def forward(self, inputs, hidden, context, context_len):
        """
        Execute RecurrentAttention.

        Args:
            inputs (int): tensor with inputs
            hidden (int): hidden state for LSTM layer
            context: context tensor from encoder
            context_len: vector of encoder sequence lengths

        Returns:
            (rnn_outputs, hidden, attn_output, attn_scores)
        """
        # set attention mask, sequences have different lengths, this mask
        # allows to include only valid elements of context in attention's
        # softmax
        self.attn.set_mask(context_len, context)

        inputs = self.dropout(inputs)
        rnn_outputs, hidden = self.rnn(inputs, hidden)
        attn_outputs, scores = self.attn(rnn_outputs, context)

        return rnn_outputs, hidden, attn_outputs, scores


class Classifier(nn.Module):
    """
    Fully-connected classifier
    """

    def __init__(self, in_features, out_features, init_weight=0.1):
        """
        Constructor for the Classifier.

        Args:
            in_features (int): number of input features
            out_features (int): number of output features (size of vocabulary)
            init_weight (float): range for the uniform initializer
        """
        super(Classifier, self).__init__()
        self.classifier = nn.Linear(in_features, out_features)
        nn.init.uniform_(self.classifier.weight.data, -init_weight, init_weight)
        nn.init.uniform_(self.classifier.bias.data, -init_weight, init_weight)

    def forward(self, x):
        """
        Execute the classifier.

        Args:
            x (torch.tensor):

        Returns:
            torch.tensor
        """
        out = self.classifier(x)
        return out


class ResidualRecurrentDecoder(nn.Module):
    """
    Decoder with Embedding, LSTM layers, attention, residual connections and
    optinal dropout.

    Attention implemented in this module is different than the attention
    discussed in the GNMT arxiv paper. In this model the output from the first
    LSTM layer of the decoder goes into the attention module, then the
    re-weighted context is concatenated with inputs to all subsequent LSTM
    layers in the decoder at the current timestep.

    Residual connections are enabled after 3rd LSTM layer, dropout is applied
    on inputs to LSTM layers.
    """

    def __init__(
        self,
        vocab_size,
        hidden_size=1024,
        num_layers=4,
        dropout=0.2,
        batch_first=False,
        embedder=None,
        init_weight=0.1,
    ):
        """
        Constructor of the ResidualRecurrentDecoder.

        Args:
            vocab_size (int): size of vocabulary
            hidden_size (int): hidden size for LSMT layers
            num_layers (int): number of LSTM layers
            dropout (float): probability of dropout (on input to LSTM layers)
            batch_first (bool): if True the model uses (batch,seq,feature)
                tensors, if false the model uses (seq, batch, feature)
            embedder (nn.Embedding): if None constructor will create new
                embedding layer
            init_weight (float): range for the uniform initializer
        """

        super(ResidualRecurrentDecoder, self).__init__()

        self.num_layers = num_layers

        self.att_rnn = RecurrentAttention(
            hidden_size,
            hidden_size,
            hidden_size,
            num_layers=1,
            batch_first=batch_first,
            dropout=dropout,
        )

        self.rnn_layers = nn.ModuleList()
        for _ in range(num_layers - 1):
            self.rnn_layers.append(
                nn.LSTM(
                    2 * hidden_size,
                    hidden_size,
                    num_layers=1,
                    bias=True,
                    batch_first=batch_first,
                )
            )

        for lstm in self.rnn_layers:
            init_lstm_(lstm, init_weight)

        if embedder is not None:
            self.embedder = embedder
        else:
            self.embedder = nn.Embedding(
                vocab_size, hidden_size, padding_idx=config.PAD
            )
            nn.init.uniform_(self.embedder.weight.data, -init_weight, init_weight)

        self.classifier = Classifier(hidden_size, vocab_size)
        self.dropout = nn.Dropout(p=dropout)

    def init_hidden(self, hidden):
        """
        Converts flattened hidden state (from sequence generator) into a tuple
        of hidden states.
        Args:
            hidden: None or flattened hidden state for decoder RNN layers

        """
        if hidden is not None:
            # per-layer chunks
            hidden = hidden.chunk(self.num_layers)
            # (h, c) chunks for LSTM layer
            hidden = tuple(i.chunk(2) for i in hidden)
        else:
            hidden = [None] * self.num_layers

        self.next_hidden = []
        return hidden

    def append_hidden(self, h):
        """
        Appends the hidden vector h to the list of internal hidden states.

        Args:
            h: hidden vector

        """
        if self.inference:
            self.next_hidden.append(h)

    def package_hidden(self):
        """
        Flattens the hidden state from all LSTM layers into one tensor (for
        the sequence generator).
        """
        if self.inference:
            hidden = torch.cat(tuple(itertools.chain(*self.next_hidden)))
        else:
            hidden = None
        return hidden

    def forward(self, inputs, context, inference=False):
        """
        Execute the decoder.

        Args:
            inputs: tensor with inputs to the decoder
            context: state of encoder, encoder sequence lengths and hidden
                state of decoder's LSTM layers
            inference: if True stores and repackages hidden state

        Returns:

        """
        self.inference = inference

        enc_context, enc_len, hidden = context
        hidden = self.init_hidden(hidden)

        x = self.embedder(inputs)

        x, h, attn, scores = self.att_rnn(x, hidden[0], enc_context, enc_len)
        self.append_hidden(h)

        x = torch.cat((x, attn), dim=2)
        x = self.dropout(x)
        x, h = self.rnn_layers[0](x, hidden[1])
        self.append_hidden(h)

        for i in range(1, len(self.rnn_layers)):
            residual = x
            x = torch.cat((x, attn), dim=2)
            x = self.dropout(x)
            x, h = self.rnn_layers[i](x, hidden[i + 1])
            self.append_hidden(h)
            x = x + residual

        x = self.classifier(x)
        hidden = self.package_hidden()

        return x, scores, [enc_context, enc_len, hidden]



PAD_TOKEN = "<pad>"
UNK_TOKEN = "<unk>"
BOS_TOKEN = "<s>"
EOS_TOKEN = "<\s>"

PAD, UNK, BOS, EOS = 0, 1, 2, 3
BPE_CODES = "bpe.32000"
VOCAB_FNAME = "vocab.bpe.32000"

TRAIN_FNAME = "train.tok.clean.bpe.32000"
VAL_FNAME = "newstest2014.tok.bpe.32000"

EXTS = (".en", ".de")



import logging

import torch
import torch.optim
import torch.utils.data
from mlbench_core.utils import AverageMeter
from mlbench_core.utils.pytorch.distributed import global_average

logger = logging.getLogger("mlbench")
LOG_EVERY_N_BATCHES = 25


def _record_train_batch_stats(
    batch_idx, loss, batch_size, tracker, num_batches_per_device_train
):
    r"""Record the stats in a training batch.

    Args:
        batch_idx (int): The id of the current batch
        loss (float): The loss of the batch
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use.
        num_batches_per_device_train (int): Number of batches per train epoch
    """
    progress = batch_idx / num_batches_per_device_train
    progress += tracker.current_epoch

    log_to_api = (
        batch_idx % LOG_EVERY_N_BATCHES == 0
        or batch_idx == num_batches_per_device_train
    )

    if tracker:
        tracker.record_loss(loss, batch_size, log_to_api=log_to_api)
    status = "Epoch {:5.2f} Batch {:4}: ".format(progress, batch_idx)

    logger.info(status + str(tracker))


class GNMTTrainer:
    def __init__(
        self,
        model,
        criterion,
        fp_optimizer,
        scheduler,
        translator,
        rank,
        schedule_per,
        tracker,
        metrics,
        iter_size,
    ):
        self.model = model
        self.batch_first = model.batch_first
        self.criterion = criterion
        self.epoch = 0
        self.rank = rank
        # Optimizers & Scheduler
        self.fp_optimizer = fp_optimizer

        self.schedule_per = schedule_per
        self.scheduler = scheduler
        self.device = next(model.parameters()).device

        self.translator = translator
        self.metrics = metrics
        self.iter_size = iter_size

        self.tracker = tracker

    def compute_model_output(self, src, tgt):
        src, src_length = src
        tgt, tgt_length = tgt

        if self.batch_first:
            output = self.model(src, src_length, tgt[:, :-1])
        else:
            output = self.model(src, src_length, tgt[:-1])

        return output

    def set_tracker(self, tracker):
        self.tracker = tracker

    def compute_loss(self, src, tgt, output):
        src, src_length = src
        tgt, tgt_length = tgt
        tgt = tgt.to(self.device)
        src_length = src_length.to(self.device)

        num_toks = {"tgt": int(sum(tgt_length - 1)), "src": int(sum(src_length))}

        if self.batch_first:
            tgt_labels = tgt[:, 1:]
            T, B = output.size(1), output.size(0)

        else:
            tgt_labels = tgt[1:]
            T, B = output.size(0), output.size(1)

        loss = self.criterion(output.view(T * B, -1), tgt_labels.contiguous().view(-1))

        loss_per_batch = loss.item()
        loss /= B * self.iter_size

        loss_per_token = loss_per_batch / num_toks["tgt"]
        loss_per_sentence = loss_per_batch / B

        return loss, loss_per_token, loss_per_sentence, num_toks

    def optimize(self, batch_idx, data, num_batches_per_device_train):

        # Whether to update the weights at this iteration
        update = (batch_idx % self.iter_size) == self.iter_size - 1
        if self.tracker:
            self.tracker.batch_start()

        # Clear gradients in the optimizer.
        if (batch_idx % self.iter_size) == 0:
            self.fp_optimizer.zero_grad()
            if self.tracker:
                self.tracker.record_batch_step("init")

        # Compute the output
        src, tgt = data.src, data.trg
        output = self.compute_model_output(src, tgt)

        if self.tracker:
            self.tracker.record_batch_step("fwd_pass")

        # Compute the loss
        stats = self.compute_loss(src, tgt, output)
        loss, loss_per_token, loss_per_sentence, num_toks = stats
        #

        if self.tracker:
            self.tracker.record_batch_step("comp_loss")

        # Backprop
        self.fp_optimizer.backward_loss(loss)

        if self.tracker:
            self.tracker.record_batch_step("backprop")

        # Opt step
        if update:
            self.fp_optimizer.step()

            if self.tracker:
                self.tracker.record_batch_step("opt_step")

        # Learning rate sheduler
        if update and self.schedule_per == "batch":
            self.scheduler.step()

        if self.tracker:
            self.tracker.batch_end()

        if self.batch_first:
            batch_size = output.size(0)
        else:
            batch_size = output.size(1)

        _record_train_batch_stats(
            batch_idx=batch_idx,
            loss=loss_per_token,
            batch_size=batch_size,
            tracker=self.tracker,
            num_batches_per_device_train=num_batches_per_device_train,
        )

    def _training(self):
        torch.set_grad_enabled(True)
        self.model.train()

        if self.tracker:
            self.tracker.train()

    def _eval(self):

        torch.set_grad_enabled(False)
        self.model.eval()

        # Set tracker in validation mode
        if self.tracker:
            self.tracker.validation()
            self.tracker.validation_start()

    def train_round(
        self, train_loader, val_loader, bleu_score=False, validate_every=None
    ):
        """
        Sets model in training mode, preallocates memory and runs training on
        data provided by data_loader.
        Args:
            data_loader: Data loader

        Returns:

        """
        # Set in training mode
        self._training()

        num_batches_per_device_train = len(train_loader)

        if self.schedule_per == "epoch":
            self.scheduler.step()

        for batch_idx, data in enumerate(train_loader):
            self.optimize(batch_idx, data, num_batches_per_device_train)

            if bleu_score and (batch_idx + 1) % validate_every == 0:
                self.validation_round(val_loader)
                self._training()

    def validate(self, loader):
        losses = AverageMeter()

        # Reset metrics
        for metric in self.metrics:
            metric.reset()

        with torch.no_grad():
            for data in loader:

                # Inference
                src, trg = data.src, data.trg
                output = self.compute_model_output(src, trg)

                # Compute loss
                stats = self.compute_loss(src, trg, output)
                loss, loss_per_token, loss_per_sentence, num_toks = stats

                # Update loss
                losses.update(loss_per_token, num_toks["tgt"])

                # Update metrics
                translated, targets = self.translator.translate(src, trg)
                for metric in self.metrics:
                    metric_value = metric(loss, translated, targets)
                    size = src[0].shape[0] if self.batch_first else src[0].shape[1]

                    metric.update(metric_value, size)

        metrics_averages = {metric: metric.average().item() for metric in self.metrics}
        loss_average = global_average(losses.sum, losses.count).item()
        return metrics_averages, loss_average

    def validation_round(self, data_loader):
        """
        Sets model in eval mode, disables gradients, preallocates memory and
        runs validation on data provided by data_loader.

        :param data_loader: data loader
        """
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        tracker = self.tracker
        self._eval()
        # Gather metrics and loss average
        metrics_values, loss = self.validate(data_loader)
        if tracker:
            tracker.validation_end()

        if len(metrics_values) > 0:
            # Save metrics
            if tracker:
                for metric, value in metrics_values.items():
                    tracker.record_metric(metric, value, log_to_api=True)

                    global_metric_value = global_average(value, 1).item()

                    if self.rank == 0:
                        tracker.record_stat(
                            "global_{}".format(metric.name),
                            global_metric_value,
                            log_to_api=True,
                        )

            #
            if self.rank == 0 and tracker:
                logger.info(
                    "{} for rank {}:(best epoch {}, current epoch {}): {:.3f}".format(
                        tracker.primary_metric.name,
                        tracker.rank,
                        tracker.best_epoch,
                        tracker.current_epoch,
                        tracker.best_metric_value,
                    )
                )
        else:
            if self.rank == 0:
                logger.info("Validation loss={:.3f}".format(loss))

        if tracker:
            tracker.record_loss(loss, log_to_api=True)

            global_loss = global_average(loss, 1).item()

            if self.rank == 0:
                tracker.record_stat("global_loss", global_loss, log_to_api=True)

        return tracker.is_best() if tracker else False



import torch.nn as nn
from mlbench_core.models.pytorch.gnmt.decoder import ResidualRecurrentDecoder
from mlbench_core.models.pytorch.gnmt.encoder import ResidualRecurrentEncoder
from mlbench_core.dataset.translation.pytorch import config
from torch.nn.functional import log_softmax


class Seq2Seq(nn.Module):
    """
    Generic Seq2Seq module, with an encoder and a decoder.
    """

    def __init__(self, encoder=None, decoder=None, batch_first=False):
        """
        Constructor for the Seq2Seq module.

        Args:
            encoder (Encoder): Model encoder
            decoder (Decoder): Model decoder
            batch_first (bool): Batch as first dim
        """
        super(Seq2Seq, self).__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.batch_first = batch_first

    def encode(self, inputs, lengths):
        """
        Applies the encoder to inputs with a given input sequence lengths.

        Args:
            inputs (torch.tensor): tensor with inputs (batch, seq_len) if
                'batch_first' else (seq_len, batch)
            lengths: vector with sequence lengths (excluding padding)

        Returns:
            torch.tensor
        """
        return self.encoder(inputs, lengths)

    def decode(self, inputs, context, inference=False):
        """
        Applies the decoder to inputs, given the context from the encoder.

        Args:
            inputs (torch.tensor): tensor with inputs (batch, seq_len) if
                'batch_first' else (seq_len, batch)
            context: context from the encoder
            inference: if True inference mode, if False training mode

        Returns:
            torch.tensor
        """
        return self.decoder(inputs, context, inference)

    def generate(self, inputs, context, beam_size):
        """
        Autoregressive generator, works with SequenceGenerator class.
        Executes decoder (in inference mode), applies log_softmax and topK for
        inference with beam search decoding.

        Args:
            inputs: tensor with inputs to the decoder
            context: context from the encoder
            beam_size: beam size for the generator

        Returns:
            (words, logprobs, scores, new_context)
            words: indices of topK tokens
            logprobs: log probabilities of topK tokens
            scores: scores from the attention module (for coverage penalty)
            new_context: new decoder context, includes new hidden states for
                decoder RNN cells
        """
        logits, scores, new_context = self.decode(inputs, context, True)
        logprobs = log_softmax(logits, dim=-1)
        logprobs, words = logprobs.topk(beam_size, dim=-1)
        return words, logprobs, scores, new_context


class GNMT(Seq2Seq):
    """
    GNMT v2 model
    """

    def __init__(
        self,
        vocab_size,
        hidden_size=1024,
        num_layers=4,
        dropout=0.2,
        batch_first=False,
        share_embedding=True,
    ):
        """
        Constructor for the GNMT v2 model.

        Args:
            vocab_size (int): size of vocabulary (number of tokens)
            hidden_size (int): internal hidden size of the model
            num_layers (int): number of layers, applies to both encoder and
                decoder
            dropout (float): probability of dropout (in encoder and decoder)
            batch_first (bool): if True the model uses (batch,seq,feature)
                tensors, if false the model uses (seq, batch, feature)
            share_embedding (bool): if True embeddings are shared between
                encoder and decoder
        """

        super(GNMT, self).__init__(batch_first=batch_first)

        if share_embedding:
            embedder = nn.Embedding(vocab_size, hidden_size, padding_idx=config.PAD)
            nn.init.uniform_(embedder.weight.data, -0.1, 0.1)
        else:
            embedder = None

        self.encoder = ResidualRecurrentEncoder(
            vocab_size, hidden_size, num_layers, dropout, batch_first, embedder
        )

        self.decoder = ResidualRecurrentDecoder(
            vocab_size, hidden_size, num_layers, dropout, batch_first, embedder
        )

    def forward(self, input_encoder, input_enc_len, input_decoder):
        context = self.encode(input_encoder, input_enc_len)
        context = (context, input_enc_len, None)
        output, _, _ = self.decode(input_decoder, context)

        return output



def _add_detailed_times(result, tracker):
    compute_time = tracker.get_total_compute_time()

    if compute_time:
        result += ", Compute: {} seconds".format(compute_time)

    communication_time = tracker.get_total_communication_time()

    if communication_time:
        result += ", Communication: {} seconds".format(communication_time)

    return result


def time_to_accuracy_goal(threshold):
    def _time_to_accuracy_goal(metric_name, value, tracker):
        if metric_name != "val_global_Prec@1":
            return None
        if value >= threshold:
            duration = tracker.get_total_train_time()

            result = (
                "{0:02d}% Top 1 Validation Accuracy reached in {1:.3f} "
                "seconds".format(threshold, duration)
            )

            result = _add_detailed_times(result, tracker)

            return result

        return None

    return _time_to_accuracy_goal


def task1_time_to_accuracy_goal():
    """ Accuracy over Time target for benchmark task 1: Image classification

    Target is 80% accuracy

    Return:
        func: time_time_to_accuracy_goal with threshold = 80
    """
    return time_to_accuracy_goal(80)


def task1_time_to_accuracy_light_goal():
    """ Accuracy over Time target for benchmark task 1: Image classification
    (Light)

    Light target is 70% accuracy

    Return:
        func: time_time_to_accuracy_goal with threshold = 70
    """
    return time_to_accuracy_goal(70)


def task2_time_to_accuracy_goal():
    """Time to accuracy goal for benchmark task 2: Linear binary classifier

    Target is an accuracy of 89%

    Return:
        func: time_time_to_accuracy_goal with threshold = 89
    """
    return time_to_accuracy_goal(89)


def task2_time_to_accuracy_light_goal():
    """Time to perplexity goal for benchmark task 2: Linear binary classifier

    Target is an accuracy of 80%

    Return:
        func: time_time_to_accuracy_goal with threshold = 80
    """
    return time_to_accuracy_goal(80)


def task3_time_to_preplexity_goal(metric_name, value, tracker):
    """Time to perplexity goal for benchmark task 3: Language Modelling

    Target is a perplexity of 50

    Args:
        metric_name(str): Name of the metric to test the value for,
        only "val_Prec@1" is counted
        value (float): Metric value to check
        tracker (`obj`:mlbench_core.utils.tracker.Tracker): Tracker object
        used for the current run
    Return:
        result (str) or `None` if target is not reached
    """

    if metric_name != "val_global_Perplexity":
        return None

    if value <= 50:
        duration = tracker.get_total_train_time()
        result = "Validation perplexity of 50 reached in {0:.3f} seconds".format(
            duration
        )

        result = _add_detailed_times(result, tracker)

        return result

    return None


def task3_time_to_preplexity_light_goal(metric_name, value, tracker):
    """Time to perplexity goal for benchmark task 3: Language Modelling

    Target is a perplexity of 50

    Args:
        metric_name(str): Name of the metric to test the value for,
        only "val_Prec@1" is counted
        value (float): Metric value to check
        tracker (`obj`:mlbench_core.utils.tracker.Tracker): Tracker object
        used for the current run
    Return:
        result (str) or `None` if target is not reached
    """

    if metric_name != "val_global_Perplexity":
        return None

    if value <= 100:
        duration = tracker.get_total_train_time()
        result = "Validation perplexity of 50 reached in {0:.3f} seconds".format(
            duration
        )

        result = _add_detailed_times(result, tracker)

        return result

    return None


def task4_time_to_bleu_goal(threshold=24):
    """Time to BLEU-score goal for benchmark task 4: GNMT machine translation"""

    def _time_to_bleu_goal(metric_name, value, tracker):
        if metric_name != "val_global_BLEU-Score":
            return None

        if value >= threshold:
            duration = tracker.get_total_train_time()
            result = "Validation BLEU-Score of {0} reached in {1:.3f} seconds".format(
                threshold, duration
            )

            result = _add_detailed_times(result, tracker)

            return result

        return None

    return _time_to_bleu_goal



r"""Helper functions."""

import datetime
import itertools
import logging
import os
import random
import shutil
import socket
import time

import deprecation
import torch
from mlbench_core.api import ApiClient
from mlbench_core.utils.pytorch.topology import FCGraph
from torch import distributed as dist


class Timeit(object):
    """Training Time Tracker

    Used to track training time for timing comparison

    Args:
        cumu (float, optional): starting time in seconds. Default: ``0.0``

    Example:
        >>> t = Timeit()
        >>> [... Do some training...]
        >>> t.pause()
        >>> [... do some non-training related things ...]
        >>> t.resume()
        >>> print(t.cumu)
    """

    def __init__(self, cumu=0):
        self.t = time.time()
        self._cumu = cumu
        self._paused = False

    def pause(self):
        """ Pause Time Tracking"""
        if not self._paused:
            self._cumu += time.time() - self.t
            self.t = time.time()
            self._paused = True

    def resume(self):
        """ Resume Time Tracking"""
        if self._paused:
            self.t = time.time()
            self._paused = False

    @property
    def cumu(self):
        """ float: total tracked time in seconds"""
        return self._cumu


def maybe_range(maximum):
    """Map an integer or None to an integer iterator starting from 0 with strid 1.

    If maximum number of batches per epoch is limited, then return an finite
    iterator. Otherwise, return an iterator of infinite length.
    """
    if maximum is None:
        counter = itertools.count(0)
    else:
        counter = range(maximum)
    return counter


def update_best_runtime_metric(tracker, metric_value, metric_name):
    """Update the runtime information to config if the metric value is the best."""
    best_metric_name = "best_{}".format(metric_name)
    is_best = metric_value > tracker.best_metric_value

    if is_best:
        tracker.best_metric_name = best_metric_name
        tracker.best_metric_value = metric_value
        tracker.best_epoch = tracker.current_epoch
    return is_best, best_metric_name


def convert_dtype(dtype, obj):
    # The object should be a ``module`` or a ``tensor``
    if dtype == "fp32":
        return obj.float()
    elif dtype == "fp64":
        return obj.double()
    else:
        raise NotImplementedError("dtype {} not supported.".format(dtype))


def config_logging(logging_level="INFO", logging_file="/mlbench.log"):
    """Setup logging modules.

    A stream handler and file handler are added to default logger `mlbench`.
    """

    class RankFilter(logging.Filter):
        def filter(self, record):
            record.rank = dist.get_rank()
            return True

    logger = logging.getLogger("mlbench")
    if len(logger.handlers) >= 2:
        return

    logger.setLevel(logging_level)
    logger.addFilter(RankFilter())

    formatter = logging.Formatter(
        "%(asctime)s %(name)s %(rank)2s %(levelname)s: %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    ch = logging.StreamHandler()
    ch.setLevel(logging_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    fh = logging.FileHandler(logging_file)
    fh.setLevel(logging_level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


def config_pytorch(use_cuda=False, seed=None, cudnn_deterministic=False):
    """Config pytorch packages.

    Fix random number for packages and initialize distributed environment for pytorch.
    Setup cuda environment for pytorch.

    :param config: A global object containing specified config.
    :type config: argparse.Namespace
    """
    # Setting `cudnn.deterministic = True` will turn on
    # CUDNN deterministic setting which can slow down training considerably.
    # Unexpected behavior may also be observed from checkpoint.
    # See: https: // github.com/pytorch/examples/blob/master/imagenet/main.py
    if cudnn_deterministic:
        # cudnn.deterministic = True
        print(
            "You have chosen to seed training. "
            "This will turn on the CUDNN deterministic setting, "
            "which can slow down your training considerably! "
            "You may see unexpected behavior when restarting "
            "from checkpoints."
        )

    if seed:
        random.seed(seed)
        torch.manual_seed(seed)

    # define the graph for the computation.
    if use_cuda:
        assert torch.cuda.is_available()

    rank = dist.get_rank()
    world_size = dist.get_world_size()
    backend = dist.get_backend() if dist.is_initialized() else None
    graph = FCGraph(rank, world_size, use_cuda)

    # enable cudnn accelerator if we are using cuda.
    if use_cuda:
        graph.assigned_gpu_id()
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True

        if torch.backends.cudnn.version() is None:
            print("CUDNN not found on device.")

        print(
            "World size={}, Rank={}, hostname={}, backend={}, cuda_available={}, cuda_device={}".format(
                world_size,
                rank,
                socket.gethostname(),
                backend,
                torch.cuda.is_available(),
                torch.cuda.current_device(),
            )
        )

    return rank, world_size, graph


@deprecation.deprecated(
    deprecated_in="1.3.1",
    details="This method was moved to mlbench_core.utils.log_metrics",
)
class LogMetrics(object):
    in_cluster = os.getenv("KUBERNETES_SERVICE_HOST") is not None

    if in_cluster:
        api = ApiClient()

    @staticmethod
    def log(run_id, rank, epoch, metric_name, value, tracker=None, time=None):
        if not LogMetrics.in_cluster:
            return

        metric_name = "{} @ {}".format(metric_name, rank)

        LogMetrics.api.post_metric(
            run_id,
            metric_name,
            value,
            metadata="{{rank: {}, epoch:{}}}".format(rank, epoch),
        )

        if tracker and time:
            tracker.records.append(
                {
                    "run_id": run_id,
                    "name": metric_name,
                    "cumulative": True,
                    "date": str(datetime.datetime.now()),
                    "time": str(time),
                    "value": str(value),
                    "metadata": "{{rank: {}, epoch:{}}}".format(rank, epoch),
                }
            )


@deprecation.deprecated(
    deprecated_in="1.1.1",
    details="This method has performance implications, use"
    " mlbench_core.utils.pytorch.helpers.LogMetrics instead",
)
def log_metrics(run_id, rank, epoch, metric_name, value, tracker=None, time=None):
    """ Log metrics to mlbench master/dashboard

    Args:
        run_id (str): The id of the current run
        rank (int): The rank of the current worker
        epoch (int): The current epoch
        metric_name (str): The name of the metric to save
        value (Any): The metric value
    """
    in_cluster = os.getenv("MLBENCH_IN_DOCKER") is None

    metric_name = "{} @ {}".format(metric_name, rank)

    if in_cluster:
        api = ApiClient()
        api.post_metric(
            run_id,
            metric_name,
            value,
            metadata="{{rank: {}, epoch:{}}}".format(rank, epoch),
        )

    if tracker and time:
        tracker.records.append(
            {
                "run_id": run_id,
                "name": metric_name,
                "cumulative": True,
                "date": str(datetime.datetime.now()),
                "time": str(time),
                "value": str(value),
                "metadata": "{{rank: {}, epoch:{}}}".format(rank, epoch),
            }
        )


def config_path(ckpt_run_dir, delete_existing_ckpts=False):
    """Config the path used during the experiments."""
    if delete_existing_ckpts:
        print("Remove previous checkpoint directory : {}".format(ckpt_run_dir))
        shutil.rmtree(ckpt_run_dir, ignore_errors=True)
    os.makedirs(ckpt_run_dir, exist_ok=True)


def iterate_dataloader(
    dataloader,
    dtype,
    max_batch_per_epoch=None,
    use_cuda=False,
    transform_target_type=None,
):
    for _, (data, target) in zip(maybe_range(max_batch_per_epoch), dataloader):

        data = convert_dtype(dtype, data)
        if transform_target_type:
            target = convert_dtype(dtype, target)

        if use_cuda:
            data, target = data.cuda(), target.cuda()

        yield data, target


def maybe_cuda(module, use_cuda):
    if use_cuda:
        module.cuda()
    return module



import logging
import os
import pickle

import cv2
import lmdb
import numpy as np
import torch.utils.data
from PIL import Image
from mlbench_core.dataset.util.tools import progress_download
from tensorpack.utils.compatible_serialize import loads

_logger = logging.getLogger("mlbench")

# All available datasets
_LIBSVM_DATASETS = [
    {"name": "australian_train", "n_samples": 690, "n_features": 14, "sparse": False},
    {"name": "duke_train", "n_samples": 38, "n_features": 7129, "sparse": True},
    {"name": "duke_test", "n_samples": 4, "n_features": 7129, "sparse": True},
    {
        "name": "epsilon_train",
        "n_samples": 400000,
        "n_features": 2000,
        "sparse": False,
        "url": "https://storage.googleapis.com/mlbench-datasets/libsvm"
        "/epsilon_train.lmdb",
    },
    {
        "name": "epsilon_test",
        "n_samples": 100000,
        "n_features": 2000,
        "sparse": False,
        "url": "https://storage.googleapis.com/mlbench-datasets/libsvm"
        "/epsilon_test.lmdb",
    },
    {"name": "rcv1_train", "n_samples": 677399, "n_features": 47236, "sparse": True},
    {"name": "synthetic_dense", "n_samples": 10000, "n_features": 100, "sparse": False},
    {
        "name": "webspam_train",
        "n_samples": 350000,
        "n_features": 16609143,
        "sparse": True,
    },
]


class LMDBDataset(torch.utils.data.Dataset):
    """
    LMDB Dataset

    Args:
        root (string): Either root directory for the database files,
            or a absolute path pointing to the file.
        target_transform (callable, optional):
            A function/transform that takes in the target and transforms it.
    """

    def __init__(
        self,
        name,
        data_type,
        root,
        is_image=False,
        target_transform=None,
        download=True,
    ):

        root, self.transform = maybe_download_lmdb(name, data_type, root)
        self.root = os.path.expanduser(root)
        self.target_transform = target_transform
        self.lmdb_files = self._get_valid_lmdb_files()

        # for each class, create an LSUNClassDataset
        self.dbs = []
        for lmdb_file in self.lmdb_files:
            self.dbs.append(
                LMDBPTClass(
                    root=lmdb_file,
                    transform=self.transform,
                    target_transform=target_transform,
                    is_image=is_image,
                )
            )

        # build up indices.
        self.indices = np.cumsum([len(db) for db in self.dbs])
        self.length = self.indices[-1]

        self._get_index_zones = self._build_indices()

    def _get_valid_lmdb_files(self):
        """get valid lmdb based on given root."""
        if not self.root.endswith(".lmdb"):
            for l in os.listdir(self.root):
                if "_" in l and "-lock" not in l:
                    yield os.path.join(self.root, l)
        else:
            yield self.root

    def _build_indices(self):
        indices = self.indices
        from_to_indices = enumerate(zip(indices[:-1], indices[1:]))

        def f(x):
            if len(list(from_to_indices)) == 0:
                return 0, x

            for ind, (from_index, to_index) in from_to_indices:
                if from_index <= x < to_index:
                    return ind, x - from_index

        return f

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: Tuple (image, target)
        """
        block_index, item_index = self._get_index_zones(index)
        image, target = self.dbs[block_index][item_index]
        return image, np.array([target])

    def __len__(self):
        return self.length

    def __repr__(self):
        fmt_str = "Dataset {}\n".format(self.__class__.__name__)
        fmt_str += "    Number of datapoints: {}\n".format(self.__len__())
        fmt_str += "    Root Location: {}\n".format(self.root)
        tmp = "    Transforms (if any): "
        fmt_str += "{0}{1}\n".format(
            tmp, self.transform.__repr__().replace("\n", "\n" + " " * len(tmp))
        )
        tmp = "    Target Transforms (if any): "
        fmt_str += "{0}{1}".format(
            tmp, self.target_transform.__repr__().replace("\n", "\n" + " " * len(tmp))
        )
        return fmt_str


class LMDBPTClass(torch.utils.data.Dataset):
    """
    LMDB Dataset loader Class

    Args:
        root (string): Either root directory for the database files,
            or a absolute path pointing to the file.
        transform (callable, optional): A function/transform that
            takes in an PIL image and returns a transformed version.
            E.g, ``transforms.RandomCrop``
        target_transform (callable, optional):
            A function/transform that takes in the target and transforms it.
        is_image (bool): Whether the dataset file is an image or not
    """

    def __init__(self, root, transform=None, target_transform=None, is_image=True):
        self.root = os.path.expanduser(root)
        self.transform = transform
        self.target_transform = target_transform
        self.is_image = is_image

        # open lmdb env.
        self.env = self._open_lmdb()

        # get file stats.
        self._get_length()

        # prepare cache_file
        self._prepare_cache()

    def _open_lmdb(self):
        return lmdb.open(
            self.root,
            subdir=os.path.isdir(self.root),
            readonly=True,
            lock=False,
            readahead=False,
            map_size=1099511627776 * 2,
            max_readers=1,
            meminit=False,
        )

    def _get_length(self):
        with self.env.begin(write=False) as txn:
            self.length = txn.stat()["entries"]

            if txn.get(b"__keys__") is not None:
                self.length -= 1

    def _prepare_cache(self):
        cache_file = self.root + "_cache_"
        if os.path.isfile(cache_file):
            self.keys = pickle.load(open(cache_file, "rb"))
        else:
            with self.env.begin(write=False) as txn:
                self.keys = [key for key, _ in txn.cursor() if key != b"__keys__"]
            pickle.dump(self.keys, open(cache_file, "wb"))

    def _image_decode(self, x):
        image = cv2.imdecode(x, cv2.IMREAD_COLOR).astype("uint8")
        return Image.fromarray(image, "RGB")

    def __getitem__(self, index):
        env = self.env
        with env.begin(write=False) as txn:
            bin_file = txn.get(self.keys[index])

        image, target = loads(bin_file)
        if self.is_image:
            image = self._image_decode(image)

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)
        return image, target

    def __len__(self):
        return self.length

    def __repr__(self):
        return self.__class__.__name__ + " (" + self.root + ")"


def construct_sparse_matrix(triplet, n_features):
    from scipy.sparse import coo_matrix, csr_matrix

    data, row, col = triplet
    mat = coo_matrix((data, (row, col)), shape=(len(set(row)), n_features))
    return csr_matrix(mat)[list(set(row))]


def maybe_transform_sparse(stats):
    return (
        (lambda x: construct_sparse_matrix(x, stats["n_features"]))
        if stats["sparse"]
        else None
    )


def get_dataset_info(name):
    stats = list(filter(lambda x: x["name"] == name, _LIBSVM_DATASETS))
    assert len(stats) == 1, "{} not found.".format(name)
    return stats[0]


def maybe_download_lmdb(name, data_type, dataset_dir):
    """ Downloads the given dataset

    Args:
        name (str): Name of the dataset, one of
            `[australian, duke, epsilon, rcv1, synthetic, webspam]`
        data_type (str): One of `test` and `train`
        dataset_dir (str): Directory where to store the dataset

    Returns:
        (str, bool): path of lmdb file and flag for sparse transform
    """

    full_name = "{}_{}".format(name, data_type)
    stats = get_dataset_info(full_name)
    lmdb_path = os.path.join(dataset_dir, "{}_{}.lmdb".format(name, data_type))

    if not (os.path.exists(lmdb_path) and os.path.isfile(lmdb_path)):
        if "url" not in stats:
            raise FileNotFoundError(
                "Could not download LIBSVM dataset {}".format(full_name)
            )
        _logger.info("Downloading dataset {}".format(full_name))

        progress_download(stats["url"], dest=lmdb_path)

    return lmdb_path, maybe_transform_sparse(stats)



r"""Define loss functions."""

import tensorflow as tf


def softmax_cross_entropy_with_logits_v2_l2_regularized(
    logits, labels, l2, loss_filter_fn
):
    """Return an op for computing cross entropy with weight decay.

    The `labels` are assumed to be one-hot encoded. The loss filter function excludes some
    tensors from computing weight decay.

    Args:
        logits (:obj:`tf.Tensor`): input logits tensor.
        labels (:obj:`tf.Tensor`): input one-hot encoded tensor.
        l2 (:obj:`float`): size of weight decay
        loss_filter_fn (:obj:`callable`): filter function.

    Returns:
        :obj:`tf.Tensor`: a scalar tensor
    """
    labels = tf.cast(labels, tf.int32)
    with tf.name_scope("loss"):
        cross_entropy = tf.reduce_mean(
            tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=labels)
        )

        loss = cross_entropy + l2 * tf.add_n(
            [
                tf.nn.l2_loss(v)
                for v in tf.trainable_variables()
                if loss_filter_fn(v.name)
            ]
        )
    return loss
